import atexit
import os
import shutil
import warnings
from abc import ABC, abstractmethod, abstractproperty
from pathlib import Path
from re import compile
from tempfile import mkdtemp
from types import TracebackType
from typing import (
    Sequence, Callable, TypeAlias, Self, TypeVar, Generic, cast, ClassVar, Any,
    overload
)

from testfixtures.comparison import compare
from testfixtures.formats import Format, JSON, YAML, TOML
from testfixtures.utils import wrap
from .rmtree import rmtree

PathStrings: TypeAlias = str | Sequence[str] | Path

P = TypeVar('P', bound=str | Path)

# When Python 3.13+ becomes the minimum version, we can use PEP 696 (type parameter defaults)
# to collapse TempDirectory and TempDir into a single class with backward-compatible
# aliases. Use PEP 695 syntax: class TempDir[P: str | Path = str]: ...
class _TempDir(ABC, Generic[P]):
    """
    A class representing a temporary directory on disk.

    :param ignore: A sequence of strings containing regular expression
                   patterns that match filenames that should be
                   ignored by the :class:`TempDirectory` listing and
                   checking methods.

    :param create: If `True`, the temporary directory will be created
                   as part of class instantiation.

    :param path: If passed, this should be a string containing an
                 absolute path to use as the temporary directory. When
                 passed, :class:`TempDirectory` will not create a new
                 directory to use.

    :param encoding: A default encoding to use for :meth:`read` and
                     :meth:`write` operations when the ``encoding`` parameter
                     is not passed to those methods.

    :param cwd: If ``True``, set the current working directory to be that of the temporary directory
                when used as a decorator or context manager.
    """

    instances: ClassVar[set['_TempDir']] = set()
    atexit_setup = False
    _path: str | None = None

    @property
    @abstractmethod
    def path(self) -> P:
        """
        The absolute path where data will be stored
        """

    def __init__(
        self,
        path: str | Path | None = None,
        *,
        ignore: Sequence[str] = (),
        create: bool | None = None,
        encoding: str | None = None,
        cwd: bool = False,
        formats: Sequence[Format] = (JSON, YAML, TOML),
    ):
        self.ignore = []
        for regex in ignore:
            self.ignore.append(compile(regex))
        self._path = str(path) if path else None
        self.encoding = encoding
        self.cwd = cwd
        self.original_cwd: str | None = None
        self.dont_remove = bool(path)

        # Build extension-to-format lookup
        self.formats: dict[str, Format] = {}
        for fmt in formats:
            for suffix in fmt.suffixes:
                if suffix in self.formats:
                    raise ValueError(
                        f"Multiple formats registered for extension '{suffix}'"
                    )
                self.formats[suffix.lower()] = fmt

        if create or (path is None and create is None):
            self.create()

    @classmethod
    def atexit(cls) -> None:
        if cls.instances:
            warnings.warn(
                'TempDirectory instances not cleaned up by shutdown:\n'
                '%s' % ('\n'.join(i._path for i in cls.instances if i._path))
                )

    def create(self) -> Self:
        """
        Create a temporary directory for this instance to use if one
        has not already been created.
        """
        if self._path:
            return self
        self._path = mkdtemp()
        self.instances.add(self)
        if not self.__class__.atexit_setup:
            atexit.register(self.atexit)
            self.__class__.atexit_setup = True
        if self.cwd:
            self.original_cwd = os.getcwd()
            os.chdir(self._path)
        return self

    def cleanup(self) -> None:
        """
        Delete the temporary directory and anything in it.
        This :class:`TempDirectory` cannot be used again unless
        :meth:`create` is called.
        """
        if self.cwd and self.original_cwd:
            os.chdir(self.original_cwd)
            self.original_cwd = None
        if self._path and os.path.exists(self._path) and not self.dont_remove:
            rmtree(self._path)
            del self._path
        if self in self.instances:
            self.instances.remove(self)

    @classmethod
    def cleanup_all(cls) -> None:
        """
        Delete all temporary directories associated with all
        :class:`TempDirectory` objects.
        """
        for i in tuple(cls.instances):
            i.cleanup()

    def actual(
            self,
            path: PathStrings | None = None,
            recursive: bool = False,
            files_only: bool = False,
            followlinks: bool = False,
    ) -> list[str]:
        path = self._join(path)

        result: list[str] = []
        if recursive:
            for dirpath, dirnames, filenames in os.walk(path, followlinks=followlinks):
                dirpath = '/'.join(dirpath[len(path)+1:].split(os.sep))
                if dirpath:
                    dirpath += '/'

                for dirname in dirnames:
                    if not files_only:
                        result.append(dirpath+dirname+'/')

                for name in sorted(filenames):
                    result.append(dirpath+name)
        else:
            for n in os.listdir(path):
                result.append(n)

        filtered = []
        for result_path in sorted(result):
            ignore = False
            for regex in self.ignore:
                if regex.search(result_path):
                    ignore = True
                    break
            if ignore:
                continue
            filtered.append(result_path)
        return filtered

    def listdir(self, path: PathStrings | None = None, recursive: bool = False) -> None:
        """
        Print the contents of the specified directory.

        :param path: The path to list, which can be:

                     * `None`, indicating the root of the temporary
                       directory should be listed.

                     * A tuple of strings, indicating that the
                       elements of the tuple should be used as directory
                       names to traverse from the root of the
                       temporary directory to find the directory to be
                       listed.

                     * A forward-slash separated string, indicating
                       the directory or subdirectory that should be
                       traversed to from the temporary directory and
                       listed.

        :param recursive: If `True`, the directory specified will have
                          its subdirectories recursively listed too.
        """
        actual = self.actual(path, recursive)
        if not actual:
            print('No files or directories found.')
        for n in actual:
            print(n)

    def compare(
            self,
            expected: Sequence[str],
            path: PathStrings | None = None,
            files_only: bool = False,
            recursive: bool = True,
            followlinks: bool = False,
    ) -> None:
        """
        Compare the expected contents with the actual contents of the temporary
        directory. An :class:`AssertionError` will be raised if they are not the
        same.

        :param expected: A sequence of strings containing the paths
                         expected in the directory. These paths should
                         be forward-slash separated and relative to
                         the root of the temporary directory.

        :param path: The path to use as the root for the comparison,
                     relative to the root of the temporary directory.
                     This can either be:

                     * A tuple of strings, making up the relative path.

                     * A forward-slash separated string.

                     If it is not provided, the root of the temporary
                     directory will be used.

        :param files_only: If specified, directories will be excluded from
                           the list of actual paths used in the comparison.

        :param recursive: If ``False``, only the direct contents of
                          the directory specified by ``path`` will be included
                          in the actual contents used for comparison.

        :param followlinks: If ``True``, symlinks and hard links
                            will be followed when recursively building up
                            the actual list of directory contents.
        """

        __tracebackhide__ = True
    
        compare(expected=sorted(expected),
                actual=tuple(self.actual(
                    path, recursive, files_only, followlinks
                )),
                recursive=False)

    def _join(self, parts: PathStrings | None) -> str:
        if self._path is None:
            raise RuntimeError('Instantiated with create=False and .create() not called')

        match parts:
            case None:
                return self._path
            case str():
                parts = parts.split('/')
            case Path():
                parts = parts.relative_to(self._path).parts

        relative = os.sep.join(parts).rstrip(os.sep)
        if relative.startswith(os.sep):
            if relative.startswith(self._path):
                return relative
            raise ValueError(
                'Attempt to read or write outside the temporary Directory'
                )
        return os.path.join(self._path, relative)

    def _makedir(self, dirpath: PathStrings) -> str:
        thepath = self._join(dirpath)
        os.makedirs(thepath)
        return thepath

    @abstractmethod
    def makedir(self, dirpath: PathStrings) -> P:
        """
        Make an empty directory at the specified path within the
        temporary directory. Any intermediate subdirectories that do
        not exist will also be created.

        :param dirpath: The directory to create, which can be:

                        * A tuple of strings.

                        * A forward-slash separated string.

        :returns: The absolute path of the created directory.
        """

    def _write(
        self,
        filepath: PathStrings,
        data: str | bytes | Any,
        encoding: str | None = None,
        format: Format | None = None
    ) -> str:
        path = Path(self._join(filepath))
        path.parent.mkdir(parents=True, exist_ok=True)

        if format is not None:
            data = format.render(data)

        encoding = encoding or self.encoding
        if encoding is not None:
            if isinstance(data, bytes):
                raise TypeError('Cannot specify encoding when data is bytes')
            data = data.encode(encoding)
        elif isinstance(data, str):
            data = data.encode()

        path.write_bytes(data)
        return str(path)

    @overload
    def write(self, filepath: PathStrings, data: bytes) -> P:
        ...

    @overload
    def write(self, filepath: PathStrings, data: str, encoding: str | None = None) -> P:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, encoding: str, format: Format) -> P:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, * , format: Format) -> P:
        ...

    @abstractmethod
    def write(
        self,
        filepath: PathStrings,
        data: str | bytes | Any,
        encoding: str | None = None,
        format: Format | None = None
    ) -> P:
        """
        Write the supplied data to a file at the specified path within
        the temporary directory. Any subdirectories specified that do
        not exist will also be created.

        The file will always be written in binary mode. The data supplied must
        either be bytes or an encoding must be supplied to convert the string
        into bytes, or a format must be supplied to serialize the data.

        :param filepath: The path to the file to create, which can be:

                         * A tuple of strings.

                         * A forward-slash separated string.

        :param data:

          :class:`bytes` containing the data to be written, a :class:`str`
          if ``encoding`` has been supplied, or any Python object if ``format``
          has been supplied.

        :param encoding: The encoding to be used for converting strings to bytes.
                         Should not be passed if data is already bytes.
                         When used with ``format``, this controls the encoding of the
                         formatted string before writing to disk.

        :param format: A :class:`~testfixtures.formats.Format` instance to use for
                       serializing the data. When used with ``encoding``, the format
                       first renders the data to a string, then that string is encoded
                       with the specified encoding.

        :returns: The absolute path of the file written.
        """

    def _find_format(self, filepath: PathStrings) -> Format:
        match filepath:
            case str():
                filename = filepath
            case Path():
                filename = filepath.name
            case _:
                filename = filepath[-1]
        suffix = os.path.splitext(filename)[-1].lower()
        if suffix not in self.formats:
            raise ValueError(
                f"No format registered for extension '{suffix}'. "
                f"Supported extensions: {sorted(self.formats.keys())}"
            )
        return self.formats[suffix]

    @abstractmethod
    def dump(self, filepath: PathStrings, data: Any, format: Format | None = None) -> P:
        """
        Serialize and write data to a file, auto-detecting format from file extension.

        :param filepath: The path to write, relative to the temporary directory.
        :param data: The Python object to serialize.
        :param format: Optional format handler. If provided, uses this format instead
                       of auto-detecting from file extension.
        :returns: The full path of the file that was written.
        """

    def _clone(
        self,
        source: str | Path,
        filepath: PathStrings | None = None,
    ) -> str:
        source_path = Path(source)
        into_dir = (
            (isinstance(filepath, str) and filepath.endswith('/')) or
            (isinstance(filepath, Sequence) and not isinstance(filepath, str) and filepath[-1] == '')
        )

        if into_dir:
            # Clone INTO directory
            dest_path = self._join(filepath)
            if source_path.is_dir():
                shutil.copytree(source_path, dest_path, dirs_exist_ok=True)
                return dest_path
            os.makedirs(dest_path, exist_ok=True)
            shutil.copy2(source_path, dest_path)
            return os.path.join(dest_path, source_path.name)

        # Clone AS path
        dest_path = self._join(source_path.name if filepath is None else filepath)
        if source_path.is_dir():
            shutil.copytree(source_path, dest_path)
            return dest_path
        if parent := os.path.dirname(dest_path):
            os.makedirs(parent, exist_ok=True)
        shutil.copy2(source_path, dest_path)
        return dest_path

    @abstractmethod
    def clone(self, source: str | Path, filepath: PathStrings | None = None) -> P:
        """
        Clone a file or directory from outside the TempDir into it.

        :param source: Path to the source file or directory.
        :param filepath: Destination within TempDir. If None, uses source's name at root.
                         If ends with '/' or is a sequence ending with '', clones into
                         that directory rather than as that path.
        :returns: Absolute path of the cloned content.
        """

    def as_string(self, path: str | Sequence[str] | None = None) -> str:
        """
        Return the full path on disk that corresponds to the path
        relative to the temporary directory that is passed in.

        :param path: The path to the file to create, which can be:

                     * A tuple of strings.

                     * A forward-slash separated string.

        :returns: A string containing the absolute path.

        """
        return self._join(path)

    #: .. deprecated:: 7
    #:
    #:   Use :meth:`as_string` instead.
    getpath = as_string

    def as_path(self, path: PathStrings | None = None) -> Path:
        """
        Return the :class:`~pathlib.Path` that corresponds to the path
        relative to the temporary directory that is passed in.

        :param path: The path to the file to create, which can be:

                     * A tuple of strings.

                     * A forward-slash separated string.
        """
        return Path(self._join(path))

    def __truediv__(self, other: str) -> Path:
        return self.as_path() / other

    @overload
    def read(self, filepath: PathStrings) -> str | bytes:
        ...

    @overload
    def read(self, filepath: PathStrings, encoding: str | None) -> str | bytes:
        ...

    @overload
    def read(self, filepath: PathStrings, *, format: Format) -> Any:
        ...

    @overload
    def read(self, filepath: PathStrings, encoding: str | None, format: Format) -> Any:
        ...

    def read(
        self,
        filepath: PathStrings,
        encoding: str | None = None,
        format: Format | None = None
    ) -> str | bytes | Any:
        """
        Reads the file at the specified path within the temporary
        directory.

        The file is always read in binary mode. Bytes will be returned unless
        an encoding is supplied, in which case a unicode string of the decoded
        data will be returned, or a format is supplied, in which case the
        deserialized Python object will be returned.

        :param filepath: The path to the file to read, which can be:

                         * A tuple of strings.

                         * A forward-slash separated string.

        :param encoding: The encoding used to decode bytes from the file into a string.
                         When used with ``format``, this controls the decoding of bytes
                         from the file before parsing.

        :param format: A :class:`~testfixtures.formats.Format` instance to use for
                       deserializing the data. When used with ``encoding``, the file bytes
                       are first decoded with the specified encoding, then parsed by the format.

        :returns:

          The contents of the file as a :class:`str`, :class:`bytes`, or deserialized
          Python object depending on the parameters provided.
        """
        with open(self._join(filepath), 'rb') as f:
            data = f.read()

        encoding = encoding or self.encoding
        if encoding is not None:
            text = data.decode(encoding)
            if format is None:
                return text
            else:
                return format.parse(text)

        if format is not None:
            return format.parse(data.decode('utf-8'))

        return data

    def read_text(
        self,
        filepath: PathStrings,
        encoding: str | None = None,
        errors: str | None = None
    ) -> str:
        """
        Read file as text, mirroring pathlib.Path.read_text() behavior.

        :param filepath: The path to the file to read.
        :param encoding: Text encoding. Uses instance default encoding if None,
                         then falls back to platform default (locale encoding).
        :param errors: Error handling strategy (e.g., 'strict', 'ignore', 'replace').
        :returns: The file contents as a string.
        """
        return Path(self._join(filepath)).read_text(encoding or self.encoding, errors)

    def read_bytes(self, filepath: PathStrings) -> bytes:
        """
        Read file as bytes, mirroring pathlib.Path.read_bytes() behavior.

        :param filepath: The path to the file to read.
        :returns: The file contents as bytes.
        """
        return Path(self._join(filepath)).read_bytes()

    def parse(self, filepath: PathStrings, format: Format | None = None) -> Any:
        """
        Read and deserialize a file, auto-detecting format from file extension.

        :param filepath: The path to read, relative to the temporary directory.
        :param format: Optional format handler. If provided, uses this format instead
                       of auto-detecting from file extension.
        :returns: The deserialized Python object.
        """
        return self.read(filepath, format=format or self._find_format(filepath))

    def __enter__(self) -> Self:
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            value: BaseException | None,
            traceback: TracebackType | None,
    ) -> None:
        self.cleanup()


class TempDirectory(_TempDir[str]):

    @property
    def path(self) -> str:
        return self._path  # type: ignore[return-value]

    def makedir(self, dirpath: PathStrings) -> str:
        return self._makedir(dirpath)

    @overload
    def write(self, filepath: PathStrings, data: bytes) -> str:
        ...

    @overload
    def write(self, filepath: PathStrings, data: str, encoding: str | None = None) -> str:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, encoding: str, format: Format) -> str:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, * , format: Format) -> str:
        ...

    def write(
        self,
        filepath: PathStrings,
        data: str | bytes | Any,
        encoding: str | None = None,
        format: Format | None = None
    ) -> str:
        return self._write(filepath, data, encoding, format)

    def dump(self, filepath: PathStrings, data: Any, format: Format | None = None) -> str:
        if format is None:
            format = self._find_format(filepath)
        return self._write(filepath, data, format=format)

    def clone(self, source: str | Path, filepath: PathStrings | None = None) -> str:
        return self._clone(source, filepath)


TempDirectory.__doc__ = (
      " .. deprecated:: 11\n    Use :class:`TempDir` instead.\n\n" + (_TempDir.__doc__ or "")
  )


class TempDir(_TempDir[Path]):

    @property
    def path(self) -> Path:
        return Path(self._path)  # type: ignore[arg-type]

    def makedir(self, dirpath: PathStrings) -> Path:
        return Path(self._makedir(dirpath))

    @overload
    def write(self, filepath: PathStrings, data: bytes) -> Path:
        ...

    @overload
    def write(self, filepath: PathStrings, data: str, encoding: str | None = None) -> Path:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, encoding: str, format: Format) -> Path:
        ...

    @overload
    def write(self, filepath: PathStrings, data: Any, * , format: Format) -> Path:
        ...

    def write(
        self,
        filepath: PathStrings,
        data: str | bytes | Any,
        encoding: str | None = None,
        format: Format | None = None
    ) -> Path:
        return Path(self._write(filepath, data, encoding, format))

    def dump(self, filepath: PathStrings, data: Any, format: Format | None = None) -> Path:
        return Path(self._write(filepath, data, format=format or self._find_format(filepath)))

    def clone(self, source: str | Path, filepath: PathStrings | None = None) -> Path:
        return Path(self._clone(source, filepath))


def tempdir(
        path: str | Path | None = None,
        *,
        ignore: Sequence[str] = (),
        encoding: str | None = None,
        cwd: bool = False,
) -> Callable[[Callable], Callable]:
    """
    .. deprecated:: 11
       Use :class:`TempDir` as a context manager, or create a `pytest fixture <https://docs.pytest.org/en/stable/how-to/fixtures.html>`_.

    A decorator for making a :class:`TempDirectory` available for the
    duration of a test function.

    All arguments and parameters are passed through to the
    :class:`TempDirectory` constructor.
    """
    l = TempDirectory(path, ignore=ignore, encoding=encoding, cwd=cwd, create=False)
    return wrap(l.create, l.cleanup)



import os
import textwrap
from dataclasses import dataclass
from typing import Iterable

from sybil import Document, Region, Example
from sybil.parsers.rest.lexers import DirectiveLexer

from testfixtures import diff


@dataclass
class FileBlock:
    path: str
    content: str
    action: str


class FileParser:
    """
    A `Sybil <http://sybil.readthedocs.io>`__ parser that
    parses certain ReST sections to read and write files in the
    configured :class:`~testfixtures.TempDirectory`.

    :param name: This is the name of the :class:`~testfixtures.TempDirectory` to use
                 in the Sybil test namespace.
    """

    def __init__(self, name: str):
        self.name = name
        self.lexer = DirectiveLexer('topic', arguments='.+')

    def __call__(self, document: Document) -> Iterable[Region]:
        for region in self.lexer(document):
            options = region.lexemes.get('options')
            if options is not None:
                class_ = options.get('class')
                if class_ in ('read-file', 'write-file'):
                    lines = region.lexemes['source'].splitlines(keepends=True)
                    index = 0
                    if lines[index].strip() == '::':
                        index += 1
                    source = textwrap.dedent(''.join(lines[index:])).lstrip()
                    if source[-1] != '\n':
                        source += '\n'
                    region.parsed = FileBlock(
                        path=region.lexemes['arguments'],
                        content=source,
                        action=class_.split('-')[0]
                    )
                    region.evaluator = self.evaluate
                    yield region

    def evaluate(self, example: Example) -> str | None:
        block: FileBlock = example.parsed
        temp_directory = example.namespace[self.name]
        if block.action == 'read':
            actual = temp_directory.as_path(block.path).read_text().replace(os.linesep, '\n')
            if actual != block.content:
                return diff(
                    block.content,
                    actual,
                    'File %r, line %i:' % (example.path, example.line),
                    'Reading from "%s":' % temp_directory.as_string(block.path)
                )
        if block.action == 'write':
            temp_directory.write(block.path, block.content)
        return None



from typing import TYPE_CHECKING, Any, Iterable, Sequence

from django.db.models import Model, Field

from .comparers import _compare_mapping

if TYPE_CHECKING:
    from .comparison import CompareContext


def instance_fields(instance: Model) -> Iterable[Field]:
    opts = instance._meta
    for name in (
        'concrete_fields',
        'virtual_fields',
        'private_fields',
    ):
        fields = getattr(opts, name, None)
        if fields:
            for field in fields:
                yield field


def model_to_dict(
        instance: Model,
        exclude: Sequence[str],
        include_not_editable: bool,
) -> dict[str, Any]:
    data = {}
    for f in instance_fields(instance):
        if f.name in exclude:
            continue
        if not getattr(f, 'editable', False) and not include_not_editable:
            continue
        data[f.name] = f.value_from_object(instance)
    return data


def compare_model(
        x: Model,
        y: Model,
        context: 'CompareContext',
        ignore_fields: Sequence[str] = (),
        non_editable_fields: bool = False,
) -> str | None:
    """
    Returns an informative string describing the differences between the two
    supplied Django model instances. The way in which this comparison is
    performed can be controlled using the following parameters:

    :param ignore_fields:
      A sequence of fields to ignore during comparison, most commonly
      set to ``['id']``. By default, no fields are ignored.

    :param non_editable_fields:
      If `True`, then fields with ``editable=False`` will be included in the
      comparison. By default, these fields are ignored.
    """
    args: Any = []
    for obj in x, y:
        args.append(model_to_dict(obj, ignore_fields, non_editable_fields))
    args.append(context)
    args.append(x)
    return _compare_mapping(*args)



import re
from collections import OrderedDict
from collections.abc import Iterable
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from functools import partial as partial_type, reduce
from inspect import signature
from operator import __or__
from pathlib import Path
from pprint import pformat
from types import GeneratorType, NotImplementedType
from typing import (
    Any,
    Sequence,
    TypeVar,
    List,
    Mapping,
    Callable,
    cast,
    overload,
    Self,
    TypeAlias,
    Literal,
    Iterator,
)
from typing import _GenericAlias as GenericAlias  # type: ignore[attr-defined]

from unittest.mock import call as unittest_mock_call

from testfixtures import not_there, singleton
from testfixtures.mock import mock_call
from testfixtures.resolve import resolve
from testfixtures.utils import indent

from .comparers import *
from .comparers import _compare_mapping, _extract_attrs

# Some common types that are immutable, for optimisation purposes within CompareContext
IMMUTABLE_TYPEs = str, bytes, int, float, tuple, type(None)

# Container types whose `__eq__` delegates to their elements. When per-type
# ignore_eq is in play, we can't trust `x == y` on these because nested
# instances of an ignored type would silently leak through.
CONTAINER_TYPES = list, tuple, dict, set, frozenset

# Things that are iterable, but should not be treated as such.
UNSAFE_ITERABLES = str, bytes, dict, GenericAlias


Comparer = Callable[[Any, Any, 'CompareContext'], str | None]
Comparers: TypeAlias = dict[type, Comparer]


DEFAULT_COMPARERS: Comparers = {
    dict: compare_dict,
    set: compare_set,
    frozenset: compare_set,
    list: compare_sequence,
    tuple: compare_tuple,
    str: compare_text,
    bytes: compare_bytes,
    int: compare_simple,
    float: compare_simple,
    Decimal: compare_simple,
    GeneratorType: compare_generator,
    mock_call.__class__: compare_call,
    unittest_mock_call.__class__: compare_call,
    BaseException: compare_exception,
    BaseExceptionGroup: compare_exception_group,
    partial_type: compare_partial,
    Path: compare_path,
    datetime: compare_with_fold,
    time: compare_with_fold,
}


@dataclass
class Registry:
    comparers: dict[type, Comparer]
    all_option_names: set[str]
    options_for: dict[Comparer, set[str]]
    ignore_eq_types: set[type]
    original: "Registry | None" = None

    @staticmethod
    def _shared_mro(x: Any, y: Any) -> Iterable[type]:
        y_mro = set(type(y).__mro__)
        for class_ in type(x).__mro__:
            if class_ in y_mro:
                yield class_

    def lookup(self, x: Any, y: Any, strict: bool) -> Comparer:
        if strict and type(x) is not type(y):
            return compare_with_type

        for class_ in self._shared_mro(x, y):
            comparer = self.comparers.get(class_)
            if comparer:
                return comparer

        # fallback for iterables
        if ((isinstance(x, Iterable) and isinstance(y, Iterable)) and not
            (isinstance(x, UNSAFE_ITERABLES) or
             isinstance(y, UNSAFE_ITERABLES))):
            return compare_generator

        # special handling for Comparisons:
        if isinstance(x, Comparison) or isinstance(y, Comparison):
            return compare_simple

        return compare_object

    def __setitem__(self, key: type, value: Comparer) -> None:
        options = set(tuple(signature(value).parameters)[3:])
        self.options_for[value] = options
        self.all_option_names |= options
        self.comparers[key] = value

    @classmethod
    def initial(cls, comparers: Comparers | None = None) -> Self:
        registry = cls(
            comparers={},
            all_option_names = {'ignore_attributes'},
            options_for = {compare_object: {'ignore_attributes'}},
            ignore_eq_types = set(),
        )
        for name, value in (DEFAULT_COMPARERS if comparers is None else comparers).items():
            registry[name] = value
        return registry

    def copy(self) -> Self:
        return type(self)(
            comparers=self.comparers.copy(),
            all_option_names = self.all_option_names.copy(),
            options_for = self.options_for.copy(),
            ignore_eq_types = self.ignore_eq_types.copy(),
        )

    def overlay_with(self, comparers: Comparers) -> Self:
        registry = self.copy()
        for name, value in comparers.items():
            registry[name] = value
        return registry

    def install(self) -> Self:
        global _registry
        self.original = _registry
        _registry = self
        return self

    def uninstall(self) -> None:
        global _registry
        assert self.original is not None, "uninstall() called without install()"
        _registry = self.original
        del self.original

_registry = Registry.initial()


@contextmanager
def registry(comparers: Comparers | None = None) -> Iterator[Registry]:
    _registry = Registry.initial(comparers)
    _registry.install()
    try:
        yield _registry
    finally:
        _registry.uninstall()


@overload
def register(type_: type, comparer: Comparer) -> None: ...
@overload
def register(type_: type, comparer: Comparer, *, ignore_eq: bool) -> None: ...
@overload
def register(type_: type, *, ignore_eq: Literal[True]) -> None: ...


def register(
    type_: type,
    comparer: Comparer | None = None,
    *,
    ignore_eq: bool = False,
) -> None:
    """
    Register the supplied comparer for the specified type, and/or mark the
    type as one whose ``__eq__`` should be ignored during comparison.

    At least one of ``comparer`` or ``ignore_eq=True`` must be supplied.

    This registration is global and will be in effect from the point
    this function is called until the end of the current process.
    """
    if comparer is None and not ignore_eq:
        raise TypeError(
            "register() requires either a comparer or ignore_eq=True"
        )
    if comparer is not None:
        _registry[type_] = comparer
    if ignore_eq:
        _registry.ignore_eq_types.add(type_)


class CompareContext:
    """
    Stores the context of the current comparison in process during a call to
    :func:`testfixtures.compare`.
    """

    def __init__(
            self,
            x_label: str | None,
            y_label: str | None,
            recursive: bool = True,
            strict: bool = False,
            ignore_eq: bool | type | Iterable[type] = False,
            comparers: Comparers | None = None,
            options: dict[str, Any] | None = None,
    ):
        self._registry = _registry.overlay_with(comparers) if comparers else _registry
        if options:
            invalid = set(options) - self._registry.all_option_names
            if invalid:
                raise TypeError(
                    'The following options are not valid: ' + ', '.join(invalid)
                )
        self.x_label = x_label
        self.y_label = y_label
        self.recursive: bool = recursive
        self.strict: bool = strict
        self.ignore_eq_all: bool = False
        self.ignore_eq_types: set[type] = set(self._registry.ignore_eq_types)
        if ignore_eq is True:
            self.ignore_eq_all = True
        elif ignore_eq is False:
            pass
        elif isinstance(ignore_eq, type):
            self.ignore_eq_types.add(ignore_eq)
        else:
            self.ignore_eq_types.update(ignore_eq)
        self.options: dict[str, Any] = options or {}
        self.message: str = ''
        self.breadcrumbs: List[str] = []
        self._seen: dict[int, str] = {}

    def extract_args(self, args: tuple, x: Any, y: Any, expected: Any, actual: Any) -> List:

        possible = list[Any]()

        def append_if_specified(source: Any) -> None:
            if source is not unspecified:
                possible.append(source)

        append_if_specified(expected)
        possible.extend(args)
        append_if_specified(actual)
        append_if_specified(x)
        append_if_specified(y)

        if len(possible) != 2:
            message = 'Exactly two objects needed, you supplied:'
            if possible:
                message += ' {}'.format(possible)
            raise TypeError(message)

        return possible

    def label(self, side: Literal["x", "y"], value: Any) -> str:
        """
        Generate a labelled representation of the value for one side of a comparison.
        """
        r = str(value)
        label = getattr(self, side+'_label')
        if label:
            r += ' ('+label+')'
        return r

    def _separator(self) -> str:
        return '\n\nWhile comparing %s: ' % ''.join(self.breadcrumbs[1:])

    def _break_loops(self, obj: Any, breadcrumb: str) -> Any:
        # Don't bother with this process for simple, immutable types:
        if isinstance(obj, IMMUTABLE_TYPEs):
            return obj

        id_ = id(obj)
        breadcrumb_ = self._seen.get(id_)
        if breadcrumb_ is not None:
            return AlreadySeen(id_, obj, breadcrumb_)
        else:
            self._seen[id_] = breadcrumb
            return obj

    def qualified_equals(self, x: Any, y: Any) -> bool:
        """
        Determines if two objects are equal, taking ``strict``, ``ignore_eq`` and instances
        of :class:`~testfixtures.comparers.AlreadySeen` into account.
        """
        pair = x, y
        # When either side is an AlreadySeen wrapper for the other (same
        # underlying id), it's the same object we already handled —
        # equal by identity. Skipping __eq__ here keeps that guarantee
        # when __eq__ is broken, strict about unknown operands, or being
        # bypassed via ignore_eq.
        if any(type(o) is AlreadySeen for o in pair):
            a, b = (o.id if type(o) is AlreadySeen else id(o) for o in pair)
            if a == b:
                return True
        # AlreadySeen.__eq__ delegates to the wrapped object, so let
        # normal equality run when the wrapper is on the right.
        if type(y) is AlreadySeen:
            return x == y
        if self.strict or self.ignore_eq_all:
            return False
        # Containers delegate __eq__ to their elements, so when any
        # ignored type is in play we must block container == as well.
        types = self.ignore_eq_types
        if types and any(
            isinstance(o, CONTAINER_TYPES) or not types.isdisjoint(type(o).__mro__)
            for o in pair
        ):
            return False
        return x == y

    def call(self, comparer: Comparer, x: Any, y: Any) -> str | None:
        kw = {}
        option_names = self._registry.options_for.get(comparer)
        if option_names:
            for name in option_names:
                value = self.options.get(name, not_there)
                if value is not not_there:
                    kw[name] = value
        return comparer(x, y, self, **kw)

    def different(self, x: Any, y: Any, breadcrumb: str) -> bool | str | None:
        """
        Comparers can call this method to :ref:`hand off <custom-comparer-different>`
        comparison of elements within the objects they are currently comparing.
        """

        x = self._break_loops(x, breadcrumb)
        y = self._break_loops(y, breadcrumb)

        recursed = bool(self.breadcrumbs)
        self.breadcrumbs.append(breadcrumb)
        existing_message = self.message
        self.message = ''
        current_message = ''
        try:

            try:
                if self.qualified_equals(x, y):
                    return False
            except RecursionError:
                pass

            comparer: Comparer = self._registry.lookup(x, y, self.strict)

            result = self.call(comparer, x, y)
            specific_comparer = comparer is not compare_simple

            if result:

                if specific_comparer and recursed:
                    current_message = self._separator()

                if specific_comparer or not recursed:
                    current_message += result

                    if self.recursive:
                        current_message += self.message

            return result

        finally:
            self.message = existing_message + current_message
            self.breadcrumbs.pop()


def _resolve_lazy(source: Any) -> str:
    value = source() if callable(source) else source
    try:
        return str(value)
    except Exception:
        return safe_repr(value)


unspecified = singleton('unspecified')


def compare(
        *args: Any,
        x: Any = unspecified,
        y: Any = unspecified,
        expected: Any = unspecified,
        actual: Any = unspecified,
        prefix: str | Callable[[], str] | None = None,
        suffix: str | Callable[[], str] | None = None,
        x_label: str | None = None,
        y_label: str | None = None,
        raises: bool = True,
        recursive: bool = True,
        strict: bool = False,
        ignore_eq: bool | type | Iterable[type] = False,
        comparers: Comparers | None = None,
        **options: Any
) -> str | None:
    """
    Compare two objects, raising an :class:`AssertionError` if they are not
    the same. The :class:`AssertionError` raised will attempt to provide
    descriptions of the differences found.

    The two objects to compare can be passed either positionally or using
    explicit keyword arguments named ``x`` and ``y``, or ``expected`` and
    ``actual``, or a mixture of these.

    :param prefix: If provided, in the event of an :class:`AssertionError`
                   being raised, the prefix supplied will be prepended to the
                   message in the :class:`AssertionError`. This may be a
                   callable, in which case it will only be resolved if needed.

    :param suffix: If provided, in the event of an :class:`AssertionError`
                   being raised, the suffix supplied will be appended to the
                   message in the :class:`AssertionError`. This may be a
                   callable, in which case it will only be resolved if needed.

    :param x_label: If provided, in the event of an :class:`AssertionError`
                    being raised, the object passed as the first positional
                    argument, or ``x`` keyword argument, will be labelled
                    with this string in the message in the
                    :class:`AssertionError`.

    :param y_label: If provided, in the event of an :class:`AssertionError`
                    being raised, the object passed as the second positional
                    argument, or ``y`` keyword argument, will be labelled
                    with this string in the message in the
                    :class:`AssertionError`.

    :param raises: If ``False``, the message that would be raised in the
                   :class:`AssertionError` will be returned instead of the
                   exception being raised.

    :param recursive: If ``True``, when a difference is found in a
                      nested data structure, attempt to highlight the location
                      of the difference.

    :param strict: If ``True``, objects will only compare equal if they are
                   of the same type as well as being equal.

    :param ignore_eq: Controls when ``__eq__`` is skipped in favour of a
                      registered comparer (or the hash-equality fallback):

                      * ``True`` — skip ``__eq__`` for *every* object.
                      * ``False`` (default) — honour ``__eq__`` except for
                        types registered with ``ignore_eq=True``.
                      * a type — also skip ``__eq__`` whenever an instance of
                        that type is on either side of a comparison.
                      * an iterable of types — as above, for each type.

    :param comparers: If supplied, should be a dictionary mapping
                      types to comparer functions for those types. These will
                      be added to the comparer registry for the duration
                      of this call.

    Any other keyword parameters supplied will be passed to the functions
    that end up doing the comparison. See the
    :mod:`API documentation below <testfixtures.comparison>`
    for details of these.
    """

    __tracebackhide__ = True

    if not (expected is unspecified and actual is unspecified):
        x_label = x_label or 'expected'
        y_label = y_label or 'actual'

    context = CompareContext(x_label, y_label, recursive, strict, ignore_eq, comparers, options)
    x, y = context.extract_args(args, x, y, expected, actual)
    if not context.different(x, y, ''):
        return None

    message = context.message
    if prefix:
        message = _resolve_lazy(prefix) + ': ' + message
    if suffix:
        message += '\n' + _resolve_lazy(suffix)

    if raises:
        raise AssertionError(message)
    return message


class StatefulComparison:
    """
    A base class for stateful comparison objects.
    """

    failed: str | None = ''
    expected: Any = None
    name_attrs: Sequence[str] = ()

    def __eq__(self, other: Any) -> bool:
        return not(self != other)

    def name(self) -> str:
        name = type(self).__name__
        if self.name_attrs:
            name += '(%s)' % ', '.join('%s=%r' % (n, getattr(self, n)) for n in self.name_attrs)
        return name

    def body(self) -> str:
        return safe_pformat(self.expected)[1:-1]

    def __repr__(self) -> str:
        name = self.name()
        body = self.failed or self.body()
        prefix = '<%s%s>' % (name, self.failed and '(failed)' or '')
        if '\n' in body:
            return '\n'+prefix+'\n'+body.strip('\n')+'\n'+'</%s>' % name
        elif body:
            return prefix + body + '</>'
        return prefix


class Comparison(StatefulComparison):
    """
    These are used when you need to compare an object's type, a subset of its attributes
    or make equality checks with objects that do not natively support comparison.

    :param object_or_type: The object or class from which to create the
                           :class:`Comparison`.

    :param attribute_dict: An optional dictionary containing attributes
                           to place on the :class:`Comparison`.

    :param partial:
      If true, only the specified attributes will be checked and any extra attributes
      of the object being compared with will be ignored.

    :param attributes: Any other keyword parameters passed will placed
                       as attributes on the :class:`Comparison`.

    """

    def __init__(self,
                 object_or_type: Any,
                 attribute_dict: dict[str, Any] | None = None,
                 partial: bool = False,
                 **attributes: Any):
        self.partial = partial
        if attributes:
            if attribute_dict is None:
                attribute_dict = attributes
            else:
                attribute_dict.update(attributes)
        if isinstance(object_or_type, str):
            c = resolve(object_or_type).found
            if c is not_there:
                raise AttributeError(
                    '%r could not be resolved' % object_or_type
                )
        elif isinstance(object_or_type, type):
            c = object_or_type
        else:
            c = object_or_type.__class__
            if attribute_dict is None:
                attribute_dict = _extract_attrs(object_or_type)
        self.expected_type = c
        self.expected_attributes = attribute_dict

    def __ne__(self, other: Any) -> bool:

        if isinstance(other, AlreadySeen):
            other = other.obj

        other_type = type(other)
        if self.expected_type is not other_type:
            self.failed = f'wrong type: {other_type.__module__}.{other_type.__qualname__}'
            return True

        if self.expected_attributes is None:
            return False

        attribute_names = set(self.expected_attributes.keys())
        actual_attributes: dict[str, Any]
        if self.partial:
            actual_attributes = {}
        else:
            actual_attributes = cast(dict[str, Any], _extract_attrs(other))
            attribute_names -= set(actual_attributes)

        for name in attribute_names:
            try:
                actual_attributes[name] = getattr(other, name)
            except AttributeError:
                pass

        context = CompareContext(x_label='Comparison', y_label='actual')
        self.failed = _compare_mapping(self.expected_attributes,
                                       actual_attributes,
                                       context,
                                       obj_for_class=not_there,
                                       prefix='attributes ',
                                       breadcrumb='.%s',
                                       check_y_not_x=not self.partial)
        return bool(self.failed)

    def name(self) -> str:
        name = 'C:'
        module = getattr(self.expected_type, '__module__', None)
        if module:
            name = name + module + '.'
        name += (getattr(self.expected_type, '__name__', None) or repr(self.expected_type))
        return name

    def body(self) -> str:
        if self.expected_attributes:
            # if we're not failed, show what we will expect:
            lines = []
            for k, v in sorted(self.expected_attributes.items()):
                rv = safe_repr(v)
                if '\n' in rv:
                    rv = indent(rv)
                lines.append(f'{k}: {rv}')
            return '\n'.join(lines)
        return ''


class SequenceComparison(StatefulComparison):
    """
    An object that can be used in comparisons of expected and actual
    sequences.

    :param expected: The items expected to be in the sequence.
    :param ordered:
      If ``True``, then the items are expected to be in the order specified.
      If ``False``, they may be in any order.
      Defaults to ``True``.
    :param partial:
      If ``True``, then any keys not expected will be ignored.
      Defaults to ``False``.
    :param recursive:
      If a difference is found, recursively compare the item where
      the difference was found to highlight exactly what was different.
      Defaults to ``False``.
    """

    name_attrs: Sequence[str] = ('ordered', 'partial')

    def __init__(
            self,
            *expected: Any,
            ordered: bool = True,
            partial: bool = False,
            recursive: bool = False,
    ):
        self.expected = expected
        self.ordered = ordered
        self.partial = partial
        self.recursive = recursive
        self.checked_indices = set[int]()

    def __ne__(self, other: Any) -> bool:
        actual: list[Any]
        try:
            actual = original_actual = list(other)
        except TypeError:
            self.failed = 'bad type'
            return True
        expected = list(self.expected)
        actual = list(actual)

        matched = []
        matched_expected_indices = []
        matched_actual_indices = []

        missing_from_expected = actual
        missing_from_expected_indices = actual_indices = list(range(len(actual)))

        missing_from_actual = []
        missing_from_actual_indices = []

        start = 0
        for e_i, e in enumerate(expected):
            try:
                i = actual.index(e, start)
                a_i = actual_indices.pop(i)
            except ValueError:
                missing_from_actual.append(e)
                missing_from_actual_indices.append(e_i)
            else:
                matched.append(missing_from_expected.pop(i))
                matched_expected_indices.append(e_i)
                matched_actual_indices.append(a_i)
                self.checked_indices.add(a_i)
                if self.ordered:
                    start = i

        matches_in_order = matched_actual_indices == sorted(matched_actual_indices)
        all_matched = not (missing_from_actual or missing_from_expected)
        partial_match = self.partial and not missing_from_actual

        if (matches_in_order or not self.ordered) and (all_matched or partial_match):
            return False

        expected_indices = matched_expected_indices+missing_from_actual_indices
        actual_indices = matched_actual_indices

        if self.partial:
            # try to give a clue as to what didn't match:
            if self.recursive and self.ordered and missing_from_expected:
                actual_indices.append(missing_from_expected_indices.pop(0))
                missing_from_expected.pop(0)

            ignored = missing_from_expected
            missing_from_expected = []
        else:
            actual_indices += missing_from_expected_indices
            ignored = None

        message = []

        def add_section(name: str, content: Any) -> None:
            if content:
                message.append(name+':\n'+pformat(content))

        add_section('ignored', ignored)

        if self.ordered:
            message.append(cast(str, compare(
                expected=[self.expected[i] for i in sorted(expected_indices)],
                actual=[original_actual[i] for i in sorted(actual_indices)],
                recursive=self.recursive,
                raises=False
            )).split('\n\n', 1)[1])
        else:
            add_section('same', matched)
            add_section('in expected but not actual', missing_from_actual)
            add_section('in actual but not expected', missing_from_expected)

        self.failed = '\n\n'.join(message)
        return True


class Subset(SequenceComparison):
    """
    A shortcut for :class:`SequenceComparison` that checks if the
    specified items are present in the sequence.
    """

    name_attrs = ()

    def __init__(self, *expected: Any) -> None:
        super(Subset, self).__init__(*expected, ordered=False, partial=True)


class Permutation(SequenceComparison):
    """
    A shortcut for :class:`SequenceComparison` that checks if the set of items
    in the sequence is as expected, but without checking ordering.
    """

    def __init__(self, *expected: Any) -> None:
        super(Permutation, self).__init__(*expected, ordered=False, partial=False)


class MappingComparison(StatefulComparison):
    """
    An object that can be used in comparisons of expected and actual
    mappings.

    :param expected_mapping:
      The mapping that should be matched expressed as either a sequence of
      ``(key, value)`` tuples or a mapping.
    :param expected_items: The items that should be matched.
    :param ordered:
      If ``True``, then the keys in the mapping are expected to be in the order specified.
      Defaults to ``False``.
    :param partial:
      If ``True``, then any keys not expected will be ignored.
      Defaults to ``False``.
    :param recursive:
      If a difference is found, recursively compare the value where
      the difference was found to highlight exactly what was different.
      Defaults to ``False``.
    """

    name_attrs = ('ordered', 'partial')

    def __init__(
            self,
            *expected_mapping: tuple[Any, Any] | Mapping[Any, Any],
            ordered: bool = False,
            partial: bool = False,
            recursive: bool = False,
            **expected_items: Any,
    ):
        self.ordered = ordered
        self.partial = partial
        self.recursive = recursive

        if len(expected_mapping) == 1:
            expected = OrderedDict(*expected_mapping)
        else:
            expected = OrderedDict(expected_mapping)  # type: ignore[arg-type]
        expected.update(expected_items)

        self.expected = expected

    def body(self) -> str:
        parts = []
        text_length = 0
        for key, value in self.expected.items():
            part = f'{safe_repr(key)}: {safe_pformat(value)}'
            text_length += len(part)
            parts.append(part)
        if text_length > 60:
            sep = ',\n'
        else:
            sep = ', '
        return sep.join(parts)

    def __ne__(self, other: Any) -> bool:
        try:
            actual_keys = other.keys()
            actual_mapping = dict(other.items())
        except AttributeError:
            self.failed = 'bad type'
            return True

        expected_keys = self.expected.keys()
        expected_mapping = self.expected

        if self.partial:
            ignored_keys = set(actual_keys) - set(expected_keys)
            for key in ignored_keys:
                del actual_mapping[key]
            # preserve the order:
            actual_keys = [k for k in actual_keys if k not in ignored_keys]
        else:
            ignored_keys = None

        mapping_differences = compare(
            expected=expected_mapping,
            actual=actual_mapping,
            recursive=self.recursive,
            raises=False
        )

        if self.ordered:
            key_differences = compare(
                expected=list(expected_keys),
                actual=list(actual_keys),
                recursive=self.recursive,
                raises=False
            )
        else:
            key_differences = None

        if key_differences or mapping_differences:

            message = []

            if ignored_keys:
                message.append('ignored:\n'+pformat(sorted(ignored_keys)))

            if mapping_differences:
                message.append(mapping_differences.split('\n\n', 1)[1])

            if key_differences:
                message.append('wrong key order:\n\n'+key_differences.split('\n\n', 1)[1])

            self.failed = '\n\n'.join(message)

            return True
        return False


class StringComparison:
    """
    An object that can be used in comparisons of expected and actual
    strings where the string expected matches a pattern rather than a
    specific concrete string.

    :param regex_source: A string containing the source for a regular
                         expression that will be used whenever this
                         :class:`StringComparison` is compared with
                         any :class:`str` instance.

    :param flags: Flags passed to :func:`re.compile`.

    :param flag_names: See the :ref:`examples <stringcomparison>`.
    """
    def __init__(self, regex_source: str, flags: int | None = None, **flag_names: str):
        args: list[Any] = [regex_source]

        flags_ = []
        if flags:
            flags_.append(flags)
        flags_.extend(getattr(re, f.upper()) for f in flag_names)
        if flags_:
            args.append(reduce(__or__, flags_))

        self.re = re.compile(*args)

    def __eq__(self, other: Any) -> bool | NotImplementedType:
        if not isinstance(other, str):
            return NotImplemented
        return self.re.match(other) is not None

    def __repr__(self) -> str:
        return '<S:%s>' % self.re.pattern

    def __lt__(self, other: Any) -> bool:
        return self.re.pattern < other

    def __gt__(self, other: Any) -> bool:
        return self.re.pattern > other


class RoundComparison:
    """
    An object that can be used in comparisons of expected and actual
    numerics to a specified precision.

    :param value: numeric to be compared.

    :param precision: Number of decimal places to round to in order
                      to perform the comparison.
    """
    def __init__(self, value: float, precision: int):
        self.rounded = round(value, precision)
        self.precision = precision

    def __eq__(self, other: Any) -> bool:
        other_rounded = round(other, self.precision)
        if type(self.rounded) is not type(other_rounded):
            raise TypeError('Cannot compare %r with %r' % (self, type(other)))
        return self.rounded == other_rounded

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return '<R:%s to %i digits>' % (self.rounded, self.precision)


class RangeComparison:
    """
    An object that can be used in comparisons of orderable types to
    check that a value specified within the given range.

    :param lower_bound: the inclusive lower bound for the acceptable range.

    :param upper_bound: the inclusive upper bound for the acceptable range.
    """
    def __init__(self, lower_bound: Any, upper_bound: Any) -> None:
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __eq__(self, other: Any) -> bool:
        return self.lower_bound <= other <= self.upper_bound

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return '<Range: [%s, %s]>' % (self.lower_bound, self.upper_bound)

T = TypeVar('T')

def like(t: type[T], **attributes: Any) -> T:
    """
    Create a type-safe partial comparison for use in strictly typed code.

    This is a convenience function that creates a :class:`Comparison` with
    ``partial=True`` but is typed to return the type being compared, making it
    compatible with strict type checkers like mypy.

    :param t: The type to compare against.
    :param attributes: Keyword arguments specifying the attributes to check.
    :return: A :class:`Comparison` object typed as the input type.
    """
    return Comparison(t, attribute_dict=attributes, partial=True)  # type: ignore[return-value]


S = TypeVar("S", bound=Sequence[Any])
S_ = TypeVar("S_", bound=Sequence[Any])


@overload
def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
) -> Callable[[S], S]: ...


@overload
def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
    *,
    returns: type[S_],
) -> Callable[[S], S_]: ...


def sequence(
    partial: bool = False,
    ordered: bool = True,
    recursive: bool = True,
    *,
    returns: type[S_] | None = None,
) -> Callable[[S], S | S_]:
    """
    Create a type-safe sequence comparison with configurable partial matching
    and ordering requirements.

    This function returns a callable that wraps a sequence in a comparison object,
    making it compatible with strict type checkers.

    :param partial: If ``True``, only items in the expected sequence need to be present
                    in the actual sequence. Defaults to ``False``.
    :param ordered: If ``True``, items must appear in the same order. Defaults to ``True``.
    :param recursive: If ``True``, provide detailed recursive comparison when differences
                      are found. Defaults to ``True``.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A callable that takes a sequence and returns a comparison object typed
             as a sequence.
    """
    def maker(items: S) -> S | S_:
        return SequenceComparison(  # type: ignore[return-value]
            *items, partial=partial, ordered=ordered, recursive=recursive
        )

    return maker


@overload
def contains(
    items: S,
) -> S: ...


@overload
def contains(
    items: S,
    *,
    returns: type[S_],
) -> S_: ...


def contains(
    items: S,
    *,
    returns: type[S_] | None = None,
) -> S | S_:
    """
    Create a type-safe partial sequence comparison that ignores order.

    Checks that the specified items are present in the actual sequence, regardless
    of their order or what other items are present. This is useful when you only
    care that certain elements exist in a collection.

    :param items: The sequence of items that must be present.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A comparison object typed as a sequence.
    """
    return SequenceComparison(  # type: ignore[return-value]
        *items, ordered=False, partial=True, recursive=True
    )


@overload
def unordered(
    items: S,
) -> S: ...


@overload
def unordered(
    items: S,
    *,
    returns: type[S_],
) -> S_: ...


def unordered(
    items: S,
    *,
    returns: type[S_] | None = None,
) -> S | S_:
    """
    Create a type-safe sequence comparison that ignores order but requires all
    items to match.

    Checks that the actual sequence contains exactly the same items as specified,
    but in any order. This is useful when order doesn't matter but you want to
    ensure no extra or missing items.

    :param items: The sequence of items that must match exactly.
    :param returns: Optional type hint for the return type, used to satisfy type checkers
                    when the comparison needs to appear as a different sequence type.
    :return: A comparison object typed as a sequence.
    """
    return SequenceComparison(  # type: ignore[return-value]
        *items, ordered=False, partial=False, recursive=True
    )


try:
    from django.db.models import Model
    from .django import compare_model
except ImportError:
    pass
else:
    register(Model, compare_model, ignore_eq=True)


try:
    from pandas import DataFrame as PandasDataFrame
    from .pandas import compare_dataframe as compare_pandas_dataframe
except ImportError:
    pass
else:
    register(PandasDataFrame, compare_pandas_dataframe, ignore_eq=True)


try:
    from polars import DataFrame as PolarsDataFrame
    from .polars import compare_dataframe as compare_polars_dataframe
except ImportError:
    pass
else:
    register(PolarsDataFrame, compare_polars_dataframe, ignore_eq=True)



"""
Tools for helping to test applications that use Loguru.
"""
from typing import TYPE_CHECKING, Callable, Any, TypeAlias
from warnings import warn

from loguru import logger

from .logcapture import Entry, build_actual_extractor, AttributeSpec

if TYPE_CHECKING:
    from loguru import Record, Message
else:
    Record = dict

RecordAttributes: TypeAlias = AttributeSpec[Record]


def level_name(record: Record) -> str:
    """
    Extract level name from the logged record

    :param record: A loguru :ref:`record dict <loguru:record>`.
    """
    return record['level'].name


class LoguruSource:
    """
    A :class:`~testfixtures.logcapture.CaptureSource` for
    `loguru <https://github.com/Delgan/loguru>`_ logging,
    for use with :class:`~testfixtures.LogCapture`.

    On :meth:`~testfixtures.logcapture.CaptureSource.install` all existing loguru handlers
    are removed and replaced with a single capture handler.
    On :meth:`~testfixtures.logcapture.CaptureSource.uninstall`, the original handlers are restored.

    :param attributes:
        The sequence of attributes to return for each :ref:`record <loguru:record>` or a callable
        that extracts :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, a key of that name will
        be obtained from the :ref:`record dict <loguru:record>`. If the key is missing,
        ``None`` will be used in its place.

        If a callable is passed, it will be called with the :ref:`record <loguru:record>` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.

    :param level:
        Minimum log level to capture.  Defaults to ``0``, capture everything.

    Additional keyword arguments are forwarded to :meth:`logger.add <loguru._logger.Logger.add>`.
    """

    def __init__(
        self,
        attributes: RecordAttributes = (level_name, 'message'),
        level: int | str = 0,
        **kw: Any,
    ) -> None:
        self.attributes = attributes
        self._kw: dict[str, Any] = {'level': level, 'colorize': False, 'catch': False, **kw}
        self._collector: Callable[[Entry], None] | None = None
        self._id: int | None = None
        self._original_handlers: dict | None = None
        self._original_min_level: int | None = None
        self._compute_actual = build_actual_extractor(attributes, self.extract_field)

    def extract_field(self, raw: Record, attribute: str) -> Any:
        return raw.get(attribute)

    def write(self, message: 'Message') -> None:
        if self._collector is not None:
            record = message.record
            exc_info = record['exception']
            entry = Entry(
                raw=record,
                actual=self._compute_actual(record),
                level=record['level'].no,
                exception=exc_info.value if exc_info is not None else None,
            )
            self._collector(entry)

    def stop(self) -> None:
        if self._collector is not None:
            warn(
                'LoguruSource left installed at shutdown.\n'
                'Call uninstall() or use LogCapture as a context manager.'
            )

    def __repr__(self) -> str:
        return 'LoguruSource()'

    def install(self, collector: Callable[[Entry], None]) -> None:
        core = logger._core  # type: ignore[attr-defined]
        self._collector = collector
        self._original_handlers = dict(core.handlers)
        self._original_min_level = core.min_level
        core.handlers = {}  # hide existing sinks, don't stop them
        self._id = logger.add(self, format="{message}", **self._kw)

    def uninstall(self) -> None:
        if self._original_handlers is not None:
            self._collector = None  # suppress the shutdown warning from stop()
            core = logger._core  # type: ignore[attr-defined]
            if self._id in core.handlers:
                logger.remove(self._id)  # stop only the capture handler
            core.min_level = self._original_min_level
            core.handlers = self._original_handlers  # restore live sinks
            self._original_handlers = None
            self._id = None



"""
Tools for helping to test applications that use structlog.
"""

import logging
import sys
from typing import Any, Callable, Sequence, TypeAlias
from warnings import warn

import structlog
from structlog._log_levels import NAME_TO_LEVEL, map_method_name
from structlog.exceptions import DropEvent
from structlog.types import EventDict, Processor

from .logcapture import Entry, build_actual_extractor, AttributeSpec

EventDictAttributes: TypeAlias = AttributeSpec[EventDict]


def level_name(event_dict: EventDict) -> str:
    """
    Extract the upper-case level name from a structlog event dict as set by
    :func:`structlog.stdlib.add_log_level`.
    """
    return str(event_dict.get('level', '')).upper()


def raw(event_dict: EventDict) -> EventDict:
    """
    Return the raw structlog event dict for use as the
    :attr:`~testfixtures.logcapture.Entry.actual` value.
    """
    return event_dict


class StructlogSource:
    """
    A :class:`~testfixtures.logcapture.CaptureSource` for
    `structlog <https://www.structlog.org/>`_, for use with
    :class:`~testfixtures.LogCapture`.

    On :meth:`~testfixtures.logcapture.CaptureSource.install`, the current
    structlog configuration is saved
    replaced with a chain that captures each event.
    On :meth:`~testfixtures.logcapture.CaptureSource.uninstall`, the
    saved configuration is restored.

    :param attributes:
        The sequence of attributes to return for each event dict, or a
        callable that extracts :attr:`~testfixtures.logcapture.Entry.actual`
        from the event dict.

        If a sequence is passed, each item may be a string key into the
        event dict or a callable that receives the event dict. Missing
        string keys yield ``None``.

    :param level:
        Minimum level to capture, as a numeric level or a structlog/stdlib
        level name (case-insensitive). Defaults to ``0`` capturing everything.

    :param processors:
        :doc:`Processors <structlog:processors>` to run before the capture processor.
        The defaults ensure that the ``level`` key is present and values
        bound via :func:`structlog.contextvars.bind_contextvars` or
        :func:`structlog.contextvars.bound_contextvars` are captured.
    """

    collector: Callable[[Entry], None] | None = None
    live_processors: list[Processor] | None = None
    saved_processors: list[Processor] | None = None

    def __init__(
        self,
        attributes: EventDictAttributes = (level_name, 'event'),
        level: int | str = 0,
        processors: Sequence[Processor] = (
            structlog.stdlib.add_log_level,
            structlog.contextvars.merge_contextvars,
        ),
    ) -> None:
        self.attributes = attributes
        self.processors: tuple[Processor, ...] = tuple(processors)
        self.min_level = self.resolve_level(level)
        self.compute_actual = build_actual_extractor(attributes, self.extract_field)

    @staticmethod
    def resolve_level(level: int | str) -> int:
        match level:
            case int():
                resolved = level
            case str():
                resolved_ = NAME_TO_LEVEL.get(map_method_name(level.lower()))
                if isinstance(resolved_, int):
                    resolved = resolved_
                else:
                    raise ValueError(f'Unknown structlog level name: {level!r}')
        return resolved

    def extract_field(self, raw: EventDict, attribute: str) -> Any:
        return raw.get(attribute)

    def capture(
        self,
        logger: Any,
        method_name: str,
        event_dict: Any,
    ) -> EventDict:
        if self.collector is None:
            raise DropEvent

        level = self.resolve_level(method_name)
        if level < self.min_level:
            raise DropEvent

        exception: BaseException | None = None
        if isinstance(event_dict, dict):
            exc_info = event_dict.get('exc_info')
            if exc_info is True:
                exc_info = sys.exc_info()
            if isinstance(exc_info, tuple) and len(exc_info) == 3:
                exception = exc_info[1]
            elif isinstance(exc_info, BaseException):
                exception = exc_info

        actual = self.compute_actual(event_dict)

        self.collector(Entry(raw=event_dict, actual=actual, level=level, exception=exception))

        raise DropEvent

    def install(self, collector: Callable[[Entry], None]) -> None:
        self.collector = collector
        # Mutate the configured processors list in place rather than swapping
        # in a new one — loggers bound (or cached) before install() hold a
        # reference to this list, and replacing it would leave them on the
        # old chain. structlog.testing.capture_logs uses the same trick.
        self.live_processors = structlog.get_config()['processors']
        self.saved_processors = list(self.live_processors)
        self.live_processors[:] = [*self.processors, self.capture]
        structlog.configure(processors=self.live_processors)

    def uninstall(self) -> None:
        if self.live_processors is None or self.saved_processors is None:
            return
        live = self.live_processors
        saved = self.saved_processors
        del self.live_processors
        del self.saved_processors
        del self.collector
        current = structlog.get_config()['processors']
        if current is not live or self.capture not in current:
            warn(
                'StructlogSource was displaced from structlog configuration before uninstall.\n'
                'Avoid calling structlog.configure() or structlog.reset_defaults() '
                'inside a LogCapture block.'
            )
            current = live
        current[:] = saved
        structlog.configure(processors=current)



import shlex
from functools import wraps, partial, reduce
from io import TextIOWrapper
from itertools import chain, zip_longest
from os import PathLike
from subprocess import STDOUT, PIPE
from tempfile import TemporaryFile
from types import TracebackType
from typing import (
    Callable, List, Sequence, Tuple, Iterable, TextIO, TypeAlias, ParamSpec,
    TypeVar, Concatenate, IO, Any, Mapping, Collection, Literal, Self, Protocol
)

from .mock import Mock, call, _Call as Call
from .utils import extend_docstring

StrOrBytesPath: TypeAlias = str | bytes | PathLike
Command: TypeAlias = StrOrBytesPath | Sequence[StrOrBytesPath]
File : TypeAlias = None | int | IO[Any]


def shell_join(command: Command) -> str:
    if isinstance(command, str):
        return command
    elif isinstance(command, bytes):
        return command.decode()
    elif isinstance(command, PathLike):
        return str(command)
    elif isinstance(command, Iterable):
        quoted_parts = []
        for part in command:
            if isinstance(part, str):
                pass
            elif isinstance(part, bytes):
                part = part.decode()
            elif isinstance(part, PathLike):
                part = str(part)
            elif not isinstance(part, (str, bytes)):
                raise TypeError(f'{part!r} in {command!r} was {type(part)}, must be str')
            quoted_parts.append(shlex.quote(part))
        return " ".join(quoted_parts)
    else:
        raise TypeError(f'{command!r} was {type(command)}, must be str')

class PopenBehaviour:
    """
    An object representing the behaviour of a :class:`MockPopen` when
    simulating a particular command.
    """

    def __init__(
            self,
            stdout: bytes = b'',
            stderr: bytes = b'',
            returncode: int = 0,
            pid: int = 1234,
            poll_count: int = 3
    ):
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode
        self.pid = pid
        self.poll_count = poll_count


class CallableBehaviour(Protocol):

    def __call__(self, command: str, stdin: File) -> PopenBehaviour: ...


P = ParamSpec('P')
R = TypeVar('R')


def record(
        func: Callable[Concatenate['MockPopenInstance', P], R]
) -> Callable[Concatenate['MockPopenInstance', P], R]:
    @wraps(func)
    def recorder(self: 'MockPopenInstance', *args: P.args, **kw: P.kwargs) -> R:
        self._record((func.__name__,), *args, **kw)
        return func(self, *args, **kw)
    return recorder


class MockPopenInstance:
    """
    A mock process as returned by :class:`MockPopen`.
    """

    #: A :class:`~unittest.mock.Mock` representing the pipe into this process.
    #: This is only set if ``stdin=PIPE`` is passed the constructor.
    #: The mock records writes and closes in :attr:`MockPopen.all_calls`.
    stdin: Mock | None = None

    #: A file representing standard output from this process.
    stdout: TextIO | None = None

    #: A file representing error output from this process.
    stderr: TextIO | None = None

    # These are not types as instantiation of this class is an internal implementation detail.
    def __init__(
            self,
            mock_class: 'MockPopen',
            root_call: Call,
            args: Command,
            bufsize: int = 0,
            executable: StrOrBytesPath | None = None,
            stdin: File = None,
            stdout: File = None,
            stderr: File = None,
            preexec_fn: Callable[[], Any] | None = None,
            close_fds: bool = False,
            shell: bool = False,
            cwd: StrOrBytesPath | None = None,
            env: Mapping[str, str] | None = None,
            universal_newlines: bool = False,
            startupinfo: Any = None,
            creationflags: int = 0,
            restore_signals: bool = True,
            start_new_session: bool = False,
            pass_fds: Collection[int] = (),
            *,
            encoding: str | None = None,
            errors: str | None = None,
            text: bool | None = None,
    ) -> None:
        self.mock: Mock = Mock()
        self.class_instance_mock: Mock = mock_class.mock.Popen_instance
        #: A :func:`unittest.mock.call` representing the call made to instantiate
        #: this mock process.
        self.root_call: Call = root_call
        #: The calls made on this mock process, represented using
        #: :func:`~unittest.mock.call` instances.
        self.calls: List[Call] = []
        self.all_calls: List[Call] = mock_class.all_calls

        cmd = shell_join(args)

        behaviour: Any
        behaviour = mock_class.commands.get(cmd, mock_class.default_behaviour)
        if behaviour is None:
            raise KeyError('Nothing specified for command %r' % cmd)

        if callable(behaviour):
            behaviour = behaviour(command=cmd, stdin=stdin)

        self.behaviour: PopenBehaviour = behaviour

        stdout_value = behaviour.stdout
        stderr_value = behaviour.stderr

        if stderr == STDOUT:
            line_iterator = chain.from_iterable(zip_longest(
                stdout_value.splitlines(True),
                stderr_value.splitlines(True)
            ))
            stdout_value = b''.join(l for l in line_iterator if l)
            stderr_value = None

        self.poll_count: int = behaviour.poll_count
        for name, option, mock_value in (
            ('stdout', stdout, stdout_value),
            ('stderr', stderr, stderr_value)
        ):
            value: Any = None
            if option is PIPE:
                value = TemporaryFile()
                value.write(mock_value)
                value.flush()
                value.seek(0)
                if universal_newlines or text or encoding:
                    value = TextIOWrapper(value, encoding=encoding, errors=errors)
            setattr(self, name, value)

        if stdin == PIPE:
            self.stdin = Mock()
            for method in 'write', 'close':
                record_writes = partial(self._record, ('stdin', method))
                getattr(self.stdin, method).side_effect = record_writes

        self.pid: int = behaviour.pid
        #: The return code of this mock process.
        self.returncode: int | None = None
        self.args: Command = args

    def _record(self, names: Sequence[str], *args: Any, **kw: Any) -> None:
        for mock in self.class_instance_mock, self.mock:
            reduce(getattr, names, mock)(*args, **kw)
        for base_call, store in (
            (call, self.calls),
            (self.root_call, self.all_calls)
        ):
            store.append(reduce(getattr, names, base_call)(*args, **kw))

    def __enter__(self) -> Self:
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            value: BaseException | None,
            traceback: TracebackType | None,
    ) -> None:
        self.wait()
        for stream in self.stdout, self.stderr:
            if stream:
                stream.close()

    @record
    def wait(self, timeout: float | None = None) -> int:
        "Simulate calls to :meth:`subprocess.Popen.wait`"
        self.returncode = self.behaviour.returncode
        return self.returncode

    @record
    def communicate(
            self, input: str | bytes | None = None, timeout: float | None = None
    ) -> Tuple[str | bytes | None, str | bytes | None]:
        "Simulate calls to :meth:`subprocess.Popen.communicate`"
        self.returncode = self.behaviour.returncode
        stdout = None if self.stdout is None else self.stdout.read()
        stderr = None if self.stderr is None else self.stderr.read()
        return stdout, stderr

    @record
    def poll(self) -> int | None:
        "Simulate calls to :meth:`subprocess.Popen.poll`"
        while self.poll_count and self.returncode is None:
            self.poll_count -= 1
            return None
        # This call to wait() is NOT how poll() behaves in reality.
        # poll() NEVER sets the returncode.
        # The returncode is *only* ever set by process completion.
        # The following is an artifact of the fixture's implementation.
        self.returncode = self.behaviour.returncode
        return self.returncode

    @record
    def send_signal(self, signal: int) -> None:
        "Simulate calls to :meth:`subprocess.Popen.send_signal`"
        pass

    @record
    def terminate(self) -> None:
        "Simulate calls to :meth:`subprocess.Popen.terminate`"
        pass

    @record
    def kill(self) -> None:
        "Simulate calls to :meth:`subprocess.Popen.kill`"
        pass


class MockPopen:
    """
    A specialised mock for testing use of :class:`subprocess.Popen`.
    An instance of this class can be used in place of the
    :class:`subprocess.Popen` and is often inserted where it's needed using
    :func:`unittest.mock.patch` or a :class:`~testfixtures.Replacer`.
    """

    default_behaviour: PopenBehaviour | CallableBehaviour | None = None

    def __init__(self) -> None:
        self.commands: dict[str, PopenBehaviour | CallableBehaviour] = {}
        self.mock: Mock = Mock()
        #: All calls made using this mock and the objects it returns, represented using
        #: :func:`~unittest.mock.call` instances.
        self.all_calls: List[Call] = []

    def _resolve_behaviour(
            self,
            stdout: bytes,
            stderr: bytes,
            returncode: int,
            pid: int,
            poll_count: int,
            behaviour: PopenBehaviour | CallableBehaviour | None
    ) -> PopenBehaviour | CallableBehaviour:
        if behaviour is None:
            return PopenBehaviour(
                stdout, stderr, returncode, pid, poll_count
            )
        else:
            return behaviour

    def set_command(
            self,
            command: str,
            stdout: bytes = b'',
            stderr: bytes = b'',
            returncode: int = 0,
            pid: int = 1234,
            poll_count: int = 3,
            behaviour: PopenBehaviour | CallableBehaviour | None = None
    ) -> None:
        """
        Set the behaviour of this mock when it is used to simulate the
        specified command.

        :param command: A :class:`str` representing the command to be simulated.
        """
        self.commands[shell_join(command)] = self._resolve_behaviour(
            stdout, stderr, returncode, pid, poll_count, behaviour
        )

    def set_default(
            self,
            stdout: bytes = b'',
            stderr: bytes = b'',
            returncode: int =0,
            pid: int = 1234,
            poll_count: int = 3,
            behaviour: PopenBehaviour | CallableBehaviour | None = None
    ) -> None:
        """
        Set the behaviour of this mock when it is used to simulate commands
        that have no explicit behavior specified using
        :meth:`~MockPopen.set_command`.
        """
        self.default_behaviour = self._resolve_behaviour(
            stdout, stderr, returncode, pid, poll_count, behaviour
        )

    def __call__(self, *args: Any, **kw: Any) -> MockPopenInstance:
        self.mock.Popen(*args, **kw)
        root_call = call.Popen(*args, **kw)
        self.all_calls.append(root_call)
        return MockPopenInstance(self, root_call, *args, **kw)


set_command_params = """
:param stdout:
    :class:`bytes` representing the simulated content written by the process
    to the stdout pipe.
:param stderr:
    :class:`bytes` representing the simulated content written by the process
    to the stderr pipe.
:param returncode:
    An integer representing the return code of the simulated process.
:param pid:
    An integer representing the process identifier of the simulated
    process. This is useful if you have code the prints out the pids
    of running processes.
:param poll_count:
    Specifies the number of times :meth:`~MockPopenInstance.poll` can be
    called before :attr:`~MockPopenInstance.returncode` is set and returned
    by :meth:`~MockPopenInstance.poll`.

If supplied, ``behaviour`` must be either a :class:`PopenBehaviour`
instance or a callable that takes the ``command`` string representing
the command to be simulated and the ``stdin`` supplied when instantiating
the :class:`subprocess.Popen` with that command and should
return a :class:`PopenBehaviour` instance.
"""


# add the param docs, so we only have one copy of them!
extend_docstring(set_command_params,
                 [MockPopen.set_command, MockPopen.set_default])



# compatibility module for different python versions
import sys
from typing import Tuple

PY_VERSION: Tuple[int, int] = sys.version_info[:2]

PY_312_PLUS: bool = PY_VERSION >= (3, 12)
PY_313_PLUS: bool = PY_VERSION >= (3, 13)



import atexit
import logging
import warnings
from collections import defaultdict
from dataclasses import dataclass
from logging import LogRecord
from pprint import pformat
from types import TracebackType
from typing import (
    Any, Callable, Sequence, TypeAlias, TypeVar, List, Tuple, Self, Protocol, overload
)
from warnings import warn

from .comparison import SequenceComparison, compare
from .utils import wrap


class CaptureSource(Protocol):
    """
    Protocol that :class:`LogCapture` sources must implement.
    """

    def install(self, collector: Callable[['Entry'], None]) -> None:
        """
        Start capturing.

        :param collector: The destination for captured :attr:`~testfixtures.LogCapture.entries`.
        """
        ...

    def uninstall(self) -> None:
        """Stop capturing and restore any previous state."""
        ...



T = TypeVar('T')
AttributeSpec: TypeAlias = Sequence[str | Callable[[T], Any]] | Callable[[T], Any]


def build_actual_extractor(
    attributes: AttributeSpec[T],
    extract_field: Callable[[T, str], Any],
) -> Callable[[T], Any]:
    # Compile the attributes spec into a callable that extracts an entry's
    # ``actual`` value from a raw record. A bare string is treated as a single
    # attribute name (``str`` is itself a ``Sequence[str]``).
    if callable(attributes):
        return attributes
    if isinstance(attributes, str):
        attributes = (attributes,)
    attrs = tuple(attributes)
    if len(attrs) == 1:
        only = attrs[0]
        if callable(only):
            return only
        return lambda raw: extract_field(raw, only)
    return lambda raw: tuple(
        a(raw) if callable(a) else extract_field(raw, a) for a in attrs
    )


@dataclass
class Entry:
    """
    A captured log entry captured by a :class:`~testfixtures.logcapture.CaptureSource`.
    """

    #: The raw object delivered by the logging framework. This is a :class:`~logging.LogRecord`
    #: for standard library logging.
    raw: Any
    #: The extracted value used by :meth:`~testfixtures.LogCapture.check` and related methods.
    #: Its structure is determined by the ``attributes`` parameter of the source.
    actual: Any
    #: Numeric log level used by :meth:`~testfixtures.LogCapture.ensure_checked`, or ``None``
    #: if the source does not provide a comparable numeric level.
    level: int | None
    #: The exception associated with this entry if one was captured, otherwise ``None``.
    exception: BaseException | None = None
    #: Whether this entry has been marked as checked by a :meth:`~testfixtures.LogCapture.check`
    #: call.
    checked: bool = False

LogRecordAttributes: TypeAlias = AttributeSpec[LogRecord]


class LogCapture:
    """
    Captures log entries from one or more sources and provides methods to check what was logged.

    :param sources:
        One or more :class:`capture sources <testfixtures.logcapture.CaptureSource>`
    :param install:
        If ``True`` (the default), capturing starts immediately on construction.
        If ``False``, installation is deferred until :meth:`install` is called.
    :param recursive_check:
        If ``True``, entries are compared recursively by :meth:`check`.
    :param ensure_checks_above:
        The log level above which entries must have been checked.
        See :meth:`ensure_checked`.

    For compatibility with earlier versions, capturing only
    standard library :mod:`logging` is supported by instantiating using these parameters:

    :param names:
        A string (or tuple of strings) containing the dotted name(s) of loggers to capture.
        By default, the root logger is captured.
    :param level:
        The minimum log level to capture.  Defaults to ``1``, capturing everything.
    :param propagate:
        If specified, any captured loggers will have their propagate attribute set to the supplied
        value. This can be used to prevent propagation from a child logger to a parent logger that
        has configured handlers.
    :param attributes:
        The sequence of attributes to return for each record or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, an attribute of that name will
        be obtained from the :class:`logging.LogRecord`. If the attribute is callable, the result of
        calling it will be used as the attribute value. If an attribute is missing, ``None`` will be
        used in its place.

        If a callable is passed, it will be called with the :class:`~logging.LogRecord` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.
    """

    #: The list of :class:`entries <testfixtures.logcapture.Entry>` captured so far.
    entries: list[Entry]

    @property
    def records(self) -> List[LogRecord]:
        """
        The records captured by this :class:`LogCapture`.

        .. deprecated:: 12
            Use the :attr:`entries` attribute instead.
        """
        return [e.raw for e in self.entries if isinstance(e.raw, LogRecord)]


    instances: set['LogCapture'] = set()
    atexit_setup = False
    installed = False

    ensure_checks_above: int | None
    #: Class-level default for :meth:`ensure_checked`. Set this to apply a level
    #: threshold to all :class:`LogCapture` instances that do not override it explicitly.
    default_ensure_checks_above: int | None = None

    @overload
    def __init__(
        self,
        *sources: CaptureSource,
        install: bool = True,
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        names: str | Tuple[str | None, ...] | None = None,
        *,
        install: bool = True,
        level: int = 1,
        propagate: bool | None = None,
        attributes: LogRecordAttributes = ...,
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None: ...

    def __init__(  # type: ignore[misc]
        self,
        *args: Any,
        install: bool = True,
        level: int = 1,
        propagate: bool | None = None,
        attributes: LogRecordAttributes = (
            'name',
            'levelname',
            'getMessage',
        ),
        recursive_check: bool = False,
        ensure_checks_above: int | None = None,
    ) -> None:
        if args and hasattr(args[0], 'install'):
            self._sources = list(args)
        else:
            names: str | Tuple[str | None, ...] | None = args[0] if args else None
            if not isinstance(names, tuple):
                names = (names,)
            self._sources = [LoggingSource(attributes, level, names=names, propagate=propagate)]
        self.recursive_check = recursive_check
        if ensure_checks_above is None:
            self.ensure_checks_above = self.default_ensure_checks_above
        else:
            self.ensure_checks_above = ensure_checks_above
        self.clear()
        if install:
            self.install()

    @classmethod
    def atexit(cls) -> None:
        if cls.instances:
            warnings.warn(
                'LogCapture instances not uninstalled by shutdown, '
                'sources captured:\n'
                '%s' % ('\n'.join((', '.join(repr(s) for s in i._sources) for i in cls.instances)))
            )

    def __len__(self) -> int:
        return len(self.entries)

    def __getitem__(self, index: int) -> Any:
        return self.entries[index].actual

    def __contains__(self, what: Any) -> bool:
        for i, entry in enumerate(self.entries):
            if what == entry.actual:
                self.entries[i].checked = True
                return True
        return False

    def clear(self) -> None:
        """Clear any entries that have been captured."""
        self.entries: list[Entry] = []

    def mark_all_checked(self) -> None:
        """
        Mark all captured events as checked.
        This should be called if you have made assertions about logging
        other than through :class:`LogCapture` methods.
        """
        for entry in self.entries:
            entry.checked = True

    def ensure_checked(self, level: int | None = None) -> None:
        """
        Ensure every entry logged above the specified `level` has been checked.
        Raises an :class:`AssertionError` if this is not the case.

        :param level:
           This defaults to the level passed during :class:`LogCapture` instantiation.
        """
        threshold: int | None = level if level is not None else self.ensure_checks_above
        if threshold is None:
            return
        un_checked = []
        for entry in self.entries:
            if entry.level is not None and entry.level >= threshold and not entry.checked:
                un_checked.append(entry.actual)
        if un_checked:
            raise AssertionError(('Not asserted ERROR log(s): %s') % (pformat(un_checked)))

    def _collect_entry(self, entry: Entry) -> None:
        self.entries.append(entry)

    def install(self) -> Self | None:
        """
        Install this :class:`LogCapture`, enabling all configured sources to begin capturing.
        """
        for source in self._sources:
            source.install(self._collect_entry)
        self.instances.add(self)
        if not self.__class__.atexit_setup:
            atexit.register(self.atexit)
            self.__class__.atexit_setup = True
        return None

    def uninstall(self) -> None:
        """
        Uninstall this :class:`LogCapture`, restoring all sources to their previous state.
        """
        if self in self.instances:
            for source in self._sources:
                source.uninstall()
            self.instances.remove(self)

    @classmethod
    def uninstall_all(cls) -> None:
        "This will uninstall all existing :class:`LogCapture` objects."
        for i in tuple(cls.instances):
            i.uninstall()

    def actual(self) -> list[Any]:
        """
        The sequence of :attr:`~testfixtures.logcapture.Entry.actual` items captured.

        This can be useful for making more complex assertions about captured logging.
        The full :class:`~testfixtures.logcapture.Entry` objects captured for each logging call
        can be accessed through :attr:`entries`.
        """
        return [e.actual for e in self.entries]

    def __str__(self) -> str:
        if not self.entries:
            return 'No logging captured'
        tuples = (r if isinstance(r, tuple) else (r,) for r in self.actual())
        return '\n'.join(' '.join(str(e) for e in t) for t in tuples)

    def check(self, *expected: Any, order_matters: bool = True, raises: bool = True) -> str | None:
        """
        This will compare the captured entries with the expected
        entries provided and raise an :class:`AssertionError` if they
        do not match.

        :param expected:

          A sequence of expected :attr:`~testfixtures.logcapture.Entry.actual` items.

        :param order_matters:

          A keyword-only parameter that controls whether the order of the
          captured entries is required to match those of the expected entries.
          Defaults to ``True``.

        :param raises: If ``False``, the message that would be raised in the
                       :class:`AssertionError` will be returned instead of the
                       exception being raised.
        """
        __tracebackhide__ = True

        result = None
        if order_matters:
            result = compare(
                expected, actual=self.actual(), recursive=self.recursive_check, raises=False
            )
        else:
            actual = self.actual()
            expected_ = SequenceComparison(
                *expected, ordered=False, partial=False, recursive=self.recursive_check
            )
            if expected_ != actual:
                result = expected_.failed
        if result and raises:
            raise AssertionError(result)
        self.mark_all_checked()
        return result

    def raise_first_exception(self, start_index: int = 0) -> None:
        """
        Raise the first captured exception from ``start_index`` onwards.

        :param start_index: The index into :attr:`entries` from where to start looking.
        """
        for entry in self.entries[start_index:]:
            if entry.exception is not None:
                raise entry.exception

    def check_exception_str(self, expected: str, index: int = -1) -> None:
        """
        Check the string representation of a captured exception.

        :param expected: The expected string.
        :param index: The index into :attr:`entries` where the exception should have been captured.
        """
        compare(expected, actual=str(self.entries[index].exception))

    def check_present(
            self, *expected: Any, order_matters: bool = True, raises: bool = True
    ) -> str | None:
        """
        This will check if the captured attr:`entries` contain all of the expected
        entries provided and raise an :class:`AssertionError` if not.
        This will ignore entries that have been captured but that do not
        match those in ``expected``.

        :param expected:

          A sequence of expected :attr:`~testfixtures.logcapture.Entry.actual` items.

        :param order_matters:

          A keyword-only parameter that controls whether the order of the
          captured entries is required to match those of the expected entries.
          Defaults to ``True``.

        :param raises: If ``False``, the message that would be raised in the
                       :class:`AssertionError` will be returned instead of the
                       exception being raised.
        """
        __tracebackhide__ = True
        actual = self.actual()
        expected_ = SequenceComparison(
            *expected, ordered=order_matters, partial=True, recursive=self.recursive_check
        )
        if expected_ != actual:
            if raises:
                raise AssertionError(expected_.failed)
            return expected_.failed
        for index in expected_.checked_indices:
            self.entries[index].checked = True
        return None

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.uninstall()
        self.ensure_checked()


class LogCaptureForDecorator(LogCapture):
    def install(self) -> Self:
        LogCapture.install(self)
        self.clear()
        return self


def log_capture(*names: str, **kw: Any) -> Callable:
    """
    A decorator for making a :class:`LogCapture` installed and
    available for the duration of a test function.

    :param names: An optional sequence of names specifying the loggers
                  to be captured. If not specified, the root logger
                  will be captured.

    Keyword parameters other than ``install`` may also be supplied and will be
    passed on to the :class:`LogCapture` constructor.
    """
    l = LogCaptureForDecorator(names or None, install=False, **kw)
    return wrap(l.install, l.uninstall)


class LoggingHandler(logging.Handler):
    """Minimal handler that forwards records to a :class:`LoggingSource`."""

    def __init__(self, source: 'LoggingSource') -> None:
        super().__init__()
        self._source = source

    def __bool__(self) -> bool:
        # Some logging internals check boolean rather than identity for handlers :-(
        return True

    def emit(self, record: LogRecord) -> None:
        self._source._handle_record(record)

    def close(self) -> None:
        super().close()
        self._source._handler_closed()


class LoggingSource:
    """
    A :class:`~testfixtures.logcapture.CaptureSource` for the standard-library
    :mod:`logging` framework, for use with :class:`~testfixtures.LogCapture`.

    :param attributes:
        The sequence of attributes to return for each record or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, an attribute of that name will
        be obtained from the :class:`logging.LogRecord`. If the attribute is callable, the result of
        calling it will be used as the attribute value. If an attribute is missing, ``None`` will be
        used in its place.

        If a callable is passed, it will be called with the :class:`~logging.LogRecord` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.
    :param level:
        The minimum log level to capture.  Defaults to ``1``, capturing everything.
    :param names:
        A string (or tuple of strings) containing the dotted name(s) of loggers to capture.
        By default, the root logger is captured.
    :param propagate:
        If specified, any captured loggers will have their propagate attribute set to the supplied
        value. This can be used to prevent propagation from a child logger to a parent logger that
        has configured handlers.
    """

    def __init__(
        self,
        attributes: LogRecordAttributes = ('levelname', 'getMessage'),
        level: int = 1,
        *,
        names: Tuple[str | None, ...] = (None,),
        propagate: bool | None = None,
    ) -> None:
        self.attributes = attributes
        self.level = level
        self.names = names
        self.propagate = propagate
        self.old: dict[str, dict[str | None, Any]] = defaultdict(dict)
        self._collector: Callable[[Entry], None] | None = None
        self._handler: LoggingHandler | None = None
        self._compute_actual = build_actual_extractor(attributes, self.extract_field)

    def extract_field(self, raw: LogRecord, attribute: str) -> Any:
        value = getattr(raw, attribute, None)
        if callable(value):
            value = value()
        return value

    def __repr__(self) -> str:
        return f'LoggingSource({self.names!r})'

    def _handler_closed(self) -> None:
        if self._collector is not None:
            warn(
                f'LoggingSource closed while still capturing loggers: {self.names!r}\n'
                'Call uninstall() or use LogCapture as a context manager.'
            )

    def _handle_record(self, record: LogRecord) -> None:
        if self._collector is None:
            return
        exception = record.exc_info[1] if record.exc_info else None
        self._collector(Entry(
            raw=record,
            actual=self._compute_actual(record),
            level=record.levelno,
            exception=exception,
        ))

    def install(self, collector: Callable[[Entry], None]) -> None:
        self._collector = collector
        self._handler = LoggingHandler(self)
        for name in self.names:
            logger = logging.getLogger(name)
            self.old['levels'][name] = logger.level
            self.old['filters'][name] = logger.filters
            self.old['handlers'][name] = logger.handlers
            self.old['disabled'][name] = logger.disabled
            self.old['propagate'][name] = logger.propagate
            logger.setLevel(self.level)
            logger.filters = []
            logger.handlers = [self._handler]
            logger.disabled = False
            if self.propagate is not None:
                logger.propagate = self.propagate

    def uninstall(self) -> None:
        for name in self.names:
            logger = logging.getLogger(name)
            logger.setLevel(self.old['levels'][name])
            logger.filters = self.old['filters'][name]
            logger.handlers = self.old['handlers'][name]
            logger.disabled = self.old['disabled'][name]
            logger.propagate = self.old['propagate'][name]
        self._handler = None
        self._collector = None



import re
from collections import defaultdict
from datetime import datetime
from difflib import unified_diff
from functools import partial as partial_type
from pathlib import Path
from pprint import pformat
from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Pattern,
    Sequence,
    TypeVar,
    TYPE_CHECKING,
)

from testfixtures import not_there
from testfixtures.mock import parent_name, _Call

if TYPE_CHECKING:
    from .comparison import CompareContext


__all__ = [
    'AlreadySeen',
    'compare_bytes',
    'compare_call',
    'compare_dict',
    'compare_exception',
    'compare_exception_group',
    'compare_generator',
    'compare_object',
    'compare_partial',
    'compare_path',
    'compare_sequence',
    'compare_set',
    'compare_simple',
    'compare_text',
    'compare_tuple',
    'compare_with_fold',
    'compare_with_type',
    'diff',
    'merge_ignored_attributes',
    'safe_pformat',
    'safe_repr',
    'sorted_by_repr',
    'split_repr',
    'strip_blank_lines',
    'trailing_whitespace_re',
]


def safe_repr(obj: Any) -> str:
    """
    A fault-tolerant version of :func:`repr`.

    :exc:`KeyboardInterrupt` and :exc:`SystemExit` are not caught.
    """
    try:
        return repr(obj)
    except Exception as e:
        type_ = type(obj)
        type_name = type_.__name__
        match obj:
            case list():
                return '[' + ', '.join(safe_repr(e) for e in obj) + ']'
            case tuple():
                body = ', '.join(safe_repr(e) for e in obj)
                suffix = ',)' if len(obj) == 1 else ')'
                return '(' + body + suffix
            case dict():
                body = ', '.join(f'{safe_repr(k)}: {safe_repr(v)}' for k, v in obj.items())
                return '{' + body + '}'
            case frozenset():
                if not obj:
                    return type_name
                return type_name + '({' + ', '.join(safe_repr(e) for e in obj) + '})'
            case set():
                if not obj:
                    return type_name + '()'
                return '{' + ', '.join(safe_repr(e) for e in obj) + '}'
            case _:
                try:
                    detail = f'{type(e).__name__}: {e}'
                except:
                    detail = type(e).__name__
                return f'<unrepresentable {type_.__module__}.{type_.__qualname__}: {detail}>'


def safe_pformat(obj: Any) -> str:
    """
    A fault-tolerant version of  :func:`pprint.pformat` but tolerant.
    Falls back to :func:`safe_repr` when :func:`~pprint.pformat` fails.
    """
    try:
        return pformat(obj)
    except:
        return safe_repr(obj)


class AlreadySeen:
    """
    A marker for an object that has already been seen during a :func:`~testfixtures.compare` call.
    These are used to prevent infinite recursion during comparison.
    """

    def __init__(self, id_: int, obj: Any, breadcrumb: str) -> None:
        self.id = id_
        self.obj = obj
        self.breadcrumb = breadcrumb

    def __repr__(self) -> str:
        return f'<AlreadySeen for {safe_repr(self.obj)} at {self.breadcrumb} with id {self.id}>'

    def __eq__(self, other: Any)-> bool:
        if isinstance(other, AlreadySeen):
            return self.breadcrumb == other.breadcrumb
        else:
            return self.obj == other


def diff(x: str, y: str, x_label: str | None = '', y_label: str | None = '') -> str:
    """
    A shorthand function that uses :mod:`difflib` to return a
    string representing the differences between the two string
    arguments.

    Most useful when comparing multi-line strings.
    """
    return '\n'.join(
        unified_diff(
            x.split('\n'),
            y.split('\n'),
            x_label or 'first',
            y_label or 'second',
            lineterm='')
    )


def compare_simple(
        x: Any, y: Any, context: 'CompareContext', newline_threshold: int = 15
) -> str | None:
    """
    Returns a very simple textual difference between the two supplied objects.
    """
    if x != y:
        repr_x = safe_repr(x)
        repr_y = safe_repr(y)
        if repr_x == repr_y:
            if type(x) is not type(y):
                return compare_with_type(x, y, context)
            x_attrs = _extract_attrs(x)
            y_attrs = _extract_attrs(y)
            diff_ = None
            if not (x_attrs is None or y_attrs is None):
                diff_ = _compare_mapping(x_attrs, y_attrs, context, x,
                                         'attributes ', '.%s')
            if diff_:
                return diff_
            x_label = context.x_label or 'x'
            y_label = context.y_label or 'y'
            return f'Both {x_label} and {y_label} appear as {repr_x!r}, but are not equal!'
        labeled_x = context.label('x', repr_x)
        labeled_y = context.label('y', repr_y)
        if len(repr_x) > newline_threshold or len(repr_y) > newline_threshold:
            return f'not equal:\n{labeled_x}\n{labeled_y}'
        else:
            return f'{labeled_x} != {labeled_y}'
    return None


def _extract_attrs(obj: Any, ignore: Iterable[str] | None = None) -> dict[str, Any] | None:
    try:
        attrs = vars(obj).copy()
    except TypeError:
        attrs = None
    else:
        if isinstance(obj, BaseException):
            attrs['args'] = obj.args

    has_slots = getattr(obj, '__slots__', not_there) is not not_there
    if has_slots:
        slots = set[str]()
        for cls in type(obj).__mro__:
            slots.update(getattr(cls, '__slots__', ()))
        if slots and attrs is None:
            attrs = {}
        for n in slots:
            value = getattr(obj, n, not_there)
            if value is not not_there:
                attrs[n] = value

    if attrs is None:
        return None

    if ignore is not None:
        for attr in ignore:
            attrs.pop(attr, None)
    return attrs


def merge_ignored_attributes(
    *ignored: Iterable[str] | Mapping[type, Iterable[str]] | str | None
) -> Mapping[type, set[str]]:
    """
    Merge multiple specifications of attributes to ignore into a single mapping.

    This is particularly useful when implementing custom comparers that need to
    combine their own attribute ignores with those passed via the context.

    Each argument can be:

    - ``None``: ignored
    - A :class:`~typing.Mapping` of type to iterable of attribute names:
      attributes for specific types
    - An iterable of attribute names: applies to all types
    - A single attribute name string: applies to all types

    Returns a mapping of types to sets of attribute names to ignore, where
    :data:`~typing.Any` is used as the key for attributes that apply to all types.
    """
    final = defaultdict[type, set[str]](set)
    for i in ignored:
        if i is None:
            pass
        elif isinstance(i, Mapping):
            for type_, values in i.items():
                final[type_].update(values)
        else:
            final[Any].update(i)
    return final


def _attrs_to_ignore(
        ignore_attributes: Iterable[str] | Mapping[type, Iterable[str]], obj: Any
) -> set[str]:
    ignored = set()
    if isinstance(ignore_attributes, dict):
        ignored |= set(ignore_attributes.get(type(obj), ()))
        ignored |= set(ignore_attributes.get(Any, ()))
    else:
        ignored.update(ignore_attributes)
    return ignored


def compare_object(
        x: object,
        y: object,
        context: 'CompareContext',
        ignore_attributes: Iterable[str] | Mapping[type, Iterable[str]] = ()
) -> str | None:
    """
    Compare the two supplied objects based on their type and attributes.

    :param ignore_attributes:

       Either a sequence of strings containing attribute names to be ignored
       when comparing, or a :class:`~typing.Mapping` of type to sequence of
       strings containing attribute names to be ignored when comparing that type.

       When specified as a mapping, you can use :data:`~typing.Any` as a key to
       specify attributes that should be ignored for all types.

    """
    if type(x) is not type(y) or isinstance(x, type):
        return compare_simple(x, y, context)
    x_attrs = _extract_attrs(x, _attrs_to_ignore(ignore_attributes, x))
    y_attrs = _extract_attrs(y, _attrs_to_ignore(ignore_attributes, y))
    if x_attrs is None or y_attrs is None or not (x_attrs and y_attrs):
        return compare_simple(x, y, context)
    if not context.qualified_equals(x_attrs, y_attrs):
        return _compare_mapping(x_attrs, y_attrs, context, x,
                                'attributes ', '.%s')
    return None


def compare_exception(
        x: BaseException, y: BaseException, context: 'CompareContext'
) -> str | None:
    """
    Compare the two supplied exceptions based on their message, type and
    attributes.
    """
    if x.args != y.args:
        return compare_simple(x, y, context)
    return context.call(compare_object, x, y)


def compare_with_type(x: Any, y: Any, context: 'CompareContext') -> str:
    """
    Return a textual description of the difference between two objects
    including information about their types.
    """
    if type(x) is AlreadySeen and type(x.obj) is type(y) and x.obj == y:
        return ''
    source = locals()
    to_render = {}
    for name in 'x', 'y':
        obj = source[name]
        to_render[name] = context.label(
            name,
            '{0} ({1!r})'.format(_short_repr(obj), type(obj))
        )
    return '{x} != {y}'.format(**to_render)


def compare_sequence(
        x: Sequence, y: Sequence, context: 'CompareContext', prefix: bool = True
) -> str | None:
    """
    Returns a textual description of the differences between the two
    supplied sequences.
    """
    l_x = len(x)
    l_y = len(y)
    i = 0
    while i < l_x and i < l_y:
        if context.different(x[i], y[i], '[%i]' % i):
            break
        i += 1

    if l_x == l_y and i == l_x:
        return None

    header = 'sequence not as expected:\n\n' if prefix else ''
    same = safe_pformat(x[:i])
    x_label = context.x_label or 'first'
    y_label = context.y_label or 'second'
    return (
        f'{header}same:\n{same}\n\n'
        f'{x_label}:\n{safe_pformat(x[i:])}\n\n'
        f'{y_label}:\n{safe_pformat(y[i:])}'
    )


def compare_generator(x: Iterable, y: Iterable, context: 'CompareContext') -> str | None:
    """
    Returns a textual description of the differences between the two
    supplied generators.

    This is done by first unwinding each of the generators supplied
    into tuples and then passing those tuples to
    :func:`compare_sequence`.
    """
    x = tuple(x)
    y = tuple(y)

    if context.qualified_equals(x, y):
        return None

    return compare_sequence(x, y, context)


def compare_tuple(x: tuple, y: tuple, context: 'CompareContext') -> str | None:
    """
    Returns a textual difference between two tuples or
    :func:`collections.namedtuple` instances.

    The presence of a ``_fields`` attribute on a tuple is used to
    decide whether or not it is a :func:`~collections.namedtuple`.
    """
    x_fields = getattr(x, '_fields', None)
    y_fields = getattr(y, '_fields', None)
    if x_fields and y_fields:
        if x_fields == y_fields:
            return _compare_mapping(dict(zip(x_fields, x)),
                                    dict(zip(y_fields, y)),
                                    context,
                                    x)
        else:
            return compare_with_type(x, y, context)
    return compare_sequence(x, y, context)


def compare_dict(x: dict, y: dict, context: 'CompareContext') -> str | None:
    """
    Returns a textual description of the differences between the two
    supplied dictionaries.
    """
    return _compare_mapping(x, y, context, x)


Item = TypeVar('Item')


def sorted_by_repr(sequence: Iterable[Item]) -> List[Item]:
    return sorted(sequence, key=safe_repr)


def _compare_mapping(
        x: Mapping, y: Mapping, context: 'CompareContext', obj_for_class: Any,
        prefix: str = '', breadcrumb: str = '[%r]',
        check_y_not_x: bool = True
) -> str | None:

    x_keys = set(x.keys())
    y_keys = set(y.keys())
    x_not_y = x_keys - y_keys
    y_not_x = y_keys - x_keys
    same = []
    diffs = []
    for key in sorted_by_repr(x_keys.intersection(y_keys)):
        if context.different(x[key], y[key], breadcrumb % (key, )):
            labelled_x = context.label('x', safe_pformat(x[key]))
            labelled_y = context.label('y', safe_pformat(y[key]))
            diffs.append(f'{safe_repr(key)}: {labelled_x} != {labelled_y}')
        else:
            same.append(key)

    if not (x_not_y or (check_y_not_x and y_not_x) or diffs):
        return None

    if obj_for_class is not_there:
        lines = []
    else:
        lines = [f'{obj_for_class.__class__.__name__} not as expected:']

    if same:
        try:
            same = sorted(same)
        except TypeError:
            pass
        lines.extend(('', f'{prefix}same:', safe_repr(same)))

    x_label = context.x_label or 'first'
    y_label = context.y_label or 'second'

    if x_not_y:
        lines.extend(('', f'{prefix}in {x_label} but not {y_label}:'))
        for key in sorted_by_repr(x_not_y):
            lines.append(f'{safe_repr(key)}: {safe_pformat(x[key])}')
    if y_not_x:
        lines.extend(('', f'{prefix}in {y_label} but not {x_label}:'))
        for key in sorted_by_repr(y_not_x):
            lines.append(f'{safe_repr(key)}: {safe_pformat(y[key])}')
    if diffs:
        lines.extend(('', f"{prefix or 'values '}differ:"))
        lines.extend(diffs)
    return '\n'.join(lines)


def compare_set(x: set, y: set, context: 'CompareContext') -> str | None:
    """
    Returns a textual description of the differences between the two
    supplied sets.
    """
    x_not_y = x - y
    y_not_x = y - x
    if not (y_not_x or x_not_y):
        return None
    lines = [f'{x.__class__.__name__} not as expected:', '']
    x_label = context.x_label or 'first'
    y_label = context.y_label or 'second'
    if x_not_y:
        lines.extend((
            f'in {x_label} but not {y_label}:',
            safe_pformat(sorted_by_repr(x_not_y)),
            '',
            ))
    if y_not_x:
        lines.extend((
            f'in {y_label} but not {x_label}:',
            safe_pformat(sorted_by_repr(y_not_x)),
            '',
            ))
    return '\n'.join(lines)+'\n'


trailing_whitespace_re: Pattern = re.compile(r'\s+$', re.MULTILINE)


def strip_blank_lines(text: str) -> str:
    result = []
    for line in text.split('\n'):
        if line and not line.isspace():
            result.append(line)
    return '\n'.join(result)


def split_repr(text: str) -> str:
    parts = text.split('\n')
    for i, part in enumerate(parts[:-1]):
        parts[i] = repr(part + '\n')
    parts[-1] = repr(parts[-1])
    return '\n'.join(parts)


def compare_text(
        x: str,
        y: str,
        context: 'CompareContext',
        blanklines: bool = True,
        trailing_whitespace: bool = True,
        show_whitespace: bool = False,
) -> str | None:
    """
    Returns an informative string describing the differences between the two
    supplied strings. The way in which this comparison is performed
    can be controlled using the following parameters:

    :param blanklines: If `False`, then when comparing multi-line
                       strings, any blank lines in either argument
                       will be ignored.

    :param trailing_whitespace: If `False`, then when comparing
                                multi-line strings, trailing
                                whilespace on lines will be ignored.

    :param show_whitespace: If `True`, then whitespace characters in
                            multi-line strings will be replaced with their
                            representations.
    """
    if not trailing_whitespace:
        x = trailing_whitespace_re.sub('', x)
        y = trailing_whitespace_re.sub('', y)
    if not blanklines:
        x = strip_blank_lines(x)
        y = strip_blank_lines(y)
    if x == y:
        return None
    labelled_x = context.label('x', repr(x))
    labelled_y = context.label('y', repr(y))
    if len(x) > 10 or len(y) > 10:
        if '\n' in x or '\n' in y:
            if show_whitespace:
                x = split_repr(x)
                y = split_repr(y)
            message = '\n' + diff(x, y, context.x_label, context.y_label)
        else:
            message = '\n%s\n!=\n%s' % (labelled_x, labelled_y)
    else:
        message = labelled_x+' != '+labelled_y
    return message


def compare_bytes(x: bytes, y: bytes, context: 'CompareContext') -> str | None:
    if x == y:
        return None
    labelled_x = context.label('x', repr(x))
    labelled_y = context.label('y', repr(y))
    return '\n%s\n!=\n%s' % (labelled_x, labelled_y)


def compare_call(x: _Call, y: _Call, context: 'CompareContext') -> str | None:
    if x == y:
        return None

    def extract(call: _Call) -> tuple[str, tuple[Any], dict[str, Any]]:
        try:
            name, args, kwargs = call
        except ValueError:
            name = None
            args, kwargs = call
        return name, args, kwargs

    x_name, x_args, x_kw = extract(x)
    y_name, y_args, y_kw = extract(y)

    if x_name == y_name and x_args == y_args and x_kw == y_kw:
        return compare_call(getattr(x, parent_name), getattr(y, parent_name), context)

    if repr(x) != repr(y):
        return compare_text(repr(x), repr(y), context)

    different = (
        context.different(x_name, y_name, ' function name') or
        context.different(x_args, y_args, ' args') or
        context.different(x_kw, y_kw, ' kw')
    )
    if not different:
        return None

    return 'mock.call not as expected:'


def compare_partial(x: partial_type, y: partial_type, context: 'CompareContext') -> str | None:
    x_attrs = dict(func=x.func, args=x.args, keywords=x.keywords)
    y_attrs = dict(func=y.func, args=y.args, keywords=y.keywords)
    if x_attrs != y_attrs:
        return _compare_mapping(x_attrs, y_attrs, context, x,
                                'attributes ', '.%s')
    return None


def compare_path(x: Path, y: Path, context: 'CompareContext') -> str | None:
    return compare_text(str(x), str(y), context)


def compare_with_fold(x: datetime, y: datetime, context: 'CompareContext') -> str | None:
    if not (x == y and x.fold == y.fold):
        repr_x = repr(x)
        repr_y = repr(y)
        if repr_x == repr_y:
            repr_x += f' (fold={x.fold})'
            repr_y += f' (fold={y.fold})'
        return context.label('x', repr_x)+' != '+context.label('y', repr_y)
    return None


def _short_repr(obj: Any) -> str:
    repr_ = safe_repr(obj)
    if len(repr_) > 30:
        repr_ = repr_[:30] + '...'
    return repr_


def compare_exception_group(
    x: BaseExceptionGroup, y: BaseExceptionGroup, context: 'CompareContext'
) -> str | None:
    """
    Compare the two supplied exception groups
    """

    x_msg, x_excs = x.args
    y_msg, y_excs = y.args
    msg_different = context.different(x_msg, y_msg, 'msg')
    excs_different = context.different(x_excs, y_excs, 'excs')
    if msg_different or excs_different:
        return 'exception group not as expected:'
    return None



"""
Tools for helping to test applications that use Polars.
"""
from typing import TYPE_CHECKING

import polars as polars
from polars import DataFrame
from polars.testing import assert_frame_equal

if TYPE_CHECKING:
    from .comparison import CompareContext


def compare_dataframe(
        x: DataFrame, y: DataFrame, context: 'CompareContext'
) -> str | None:
    """
    Returns a textual description of the differences between two
    :class:`polars.DataFrame` instances, as reported by
    :func:`polars.testing.assert_frame_equal`.

    When ``strict=True`` is passed to :func:`~testfixtures.compare`,
    ``check_exact=True`` is used; otherwise polars' default tolerances apply.
    """
    try:
        assert_frame_equal(x, y, check_exact=context.strict)
    except AssertionError as e:
        return str(e)
    return None



import os
import sys
from io import StringIO
from tempfile import TemporaryFile
from typing import Self, Any, IO

from .comparison import StringComparison, compare


class OutputCapture:
    """
    A context manager for capturing output to the
    :any:`sys.stdout` and :any:`sys.stderr` streams.

    :param separate: If ``True``, ``stdout`` and ``stderr`` will be captured
                     separately and their expected values must be passed to
                     :meth:`~OutputCapture.compare`.

    :param fd: If ``True``, the underlying file descriptors will be captured,
               rather than just the attributes on :mod:`sys`. This allows
               you to capture things like subprocesses that write directly
               to the file descriptors, but is more invasive, so only use it
               when you need it.

    :param strip_whitespace:
        When ``True``, which is the default, leading and training whitespace
        is trimmed from both the expected and actual values when comparing.

    .. note:: If ``separate`` is passed as ``True``,
              :attr:`OutputCapture.captured` will be an empty string.
    """

    output: IO
    stdout: IO
    stderr: IO

    original_stdout: IO[str] | int | None = None
    original_stderr: IO[str] | int | None = None

    def __init__(self, separate: bool = False, fd: bool = False, strip_whitespace: bool = True):
        self.separate = separate
        self.fd = fd
        self.strip_whitespace = strip_whitespace

    def __enter__(self) -> Self:
        if self.fd:
            self.output = TemporaryFile()
            self.stdout = TemporaryFile()
            self.stderr = TemporaryFile()
        else:
            self.output = StringIO()
            self.stdout = StringIO()
            self.stderr = StringIO()
        self.enable()
        return self

    def __exit__(self, *args: Any) -> None:
        self.disable()

    def disable(self) -> None:
        "Disable the output capture if it is enabled."
        if self.fd:
            for original, current in (
                (self.original_stdout, sys.stdout),
                (self.original_stderr, sys.stderr),
            ):
                os.dup2(original, current.fileno())  # type: ignore[arg-type]
                os.close(original)  # type: ignore[arg-type]

        else:
            sys.stdout = self.original_stdout
            sys.stderr = self.original_stderr

    def enable(self) -> None:
        "Enable the output capture if it is disabled."
        if self.original_stdout is None:
            if self.fd:
                self.original_stdout = os.dup(sys.stdout.fileno())
                self.original_stderr = os.dup(sys.stderr.fileno())
            else:
                self.original_stdout = sys.stdout
                self.original_stderr = sys.stderr
        if self.separate:
            if self.fd:
                os.dup2(self.stdout.fileno(), sys.stdout.fileno())
                os.dup2(self.stderr.fileno(), sys.stderr.fileno())
            else:
                sys.stdout = self.stdout
                sys.stderr = self.stderr
        else:
            if self.fd:
                os.dup2(self.output.fileno(), sys.stdout.fileno())
                os.dup2(self.output.fileno(), sys.stderr.fileno())
            else:
                sys.stdout = sys.stderr = self.output

    def _read(self, stream: IO | StringIO) -> str:
        if self.fd:
            stream.seek(0)
            return stream.read().decode()  # type: ignore[union-attr]
        else:
            return stream.getvalue()  # type: ignore[union-attr]

    @property
    def captured(self) -> str:
        "A property containing any output that has been captured so far."
        return self._read(self.output)

    def compare(
            self,
            expected: str | StringComparison = '',
            stdout: str | StringComparison = '',
            stderr: str | StringComparison = '',
            raises: bool = True,
    ) -> str | None:
        """
        Compare the captured output to that expected. If the output is
        not the same, an :class:`AssertionError` will be raised.

        :param expected: A string containing the expected combined output
                         of ``stdout`` and ``stderr``.

        :param stdout: A string containing the expected output to ``stdout``.

        :param stderr: A string containing the expected output to ``stderr``.

        :param raises: If ``False``, the message that would be raised in the
                       :class:`AssertionError` will be returned instead of the
                       exception being raised.
        """
        __tracebackhide__ = True
        expected_mapping = {}
        actual_mapping = {}
        for prefix, _expected, captured in (
                ('captured', expected, self.captured),
                ('stdout', stdout, self._read(self.stdout)),
                ('stderr', stderr, self._read(self.stderr)),
        ):
            if self.strip_whitespace:
                if isinstance(_expected, str):
                    _expected = _expected.strip()
                captured = captured.strip()
            if _expected != captured:
                expected_mapping[prefix] = _expected
                actual_mapping[prefix] = captured
        if len(expected_mapping) == 1:
            return compare(expected=tuple(expected_mapping.values())[0],
                           actual=tuple(actual_mapping.values())[0],
                           raises=raises)
        return compare(expected=expected_mapping, actual=actual_mapping, raises=raises)



import re
from contextlib import contextmanager
from functools import wraps
from types import TracebackType
from typing import Callable, TypeAlias, Iterator, Self, ParamSpec, TypeVar, Generic, cast, overload

from testfixtures import diff, compare, not_there, singleton
from .comparers import split_repr


param_docs = """

    :param exception: This can be one of the following:

                      * Not passed, indicating that an exception must be
                        raised, but the type is unimportant.

                      * ``None``, indicating that no exception should be raised.
                        This is useful for parameterised tests where the parameter
                        may be either an exception or ``None``.

                      * An exception class, indicating that the type
                        of the exception is important but not the
                        parameters it is created with.

                      * An exception instance, indicating that an
                        exception exactly matching the one supplied
                        should be raised.

    :param unless: Can be passed a boolean that, when ``True`` indicates that
                   no exception is expected. This is useful when checking
                   that exceptions are only raised on certain versions of
                   Python.
"""


class NoException(AssertionError):
    """
    A marker class indicating no exception has been raised.

    .. currentmodule:: testfixtures

    :attr:`ShouldRaise.raised` is set to an instance of this class unless an
    exception has otherwise been seen.
    """

    def __init__(self) -> None:
        super().__init__('No exception raised!')


E = TypeVar("E", bound=BaseException)


class ShouldRaise(Generic[E]):
    __doc__ = """
    This context manager is used to assert that an exception is raised
    within the context it is managing.
    """ + param_docs

    #: The exception captured by the context manager.
    #: Can be used to inspect specific attributes of the exception.
    raised: E = NoException()  # type: ignore[assignment]
    exception: E | type[E] | None
    expected: bool
    match: str | re.Pattern[str] | None

    @overload
    def __init__(self, exception: type[E], *, match: str | re.Pattern[str] | None = None, unless: bool | None = False) -> None: ...
    @overload
    def __init__(self, exception: singleton = not_there, *, match: str | re.Pattern[str] | None = None, unless: bool | None = False) -> None: ...
    @overload
    def __init__(self, exception: E | None, *, match: None = None, unless: bool | None = False) -> None: ...

    def __init__(
            self,
            exception: E | type[E] | None | singleton = not_there,
            *,
            match: str | re.Pattern[str] | None = None,
            unless: bool | None = False,
    ) -> None:
        if match is not None and (exception is None or (exception is not not_there and not isinstance(exception, type))):
            raise TypeError('match can only be used when an exception type is provided')
        self.exception = None if exception in (None, not_there) else cast('E | type[E]', exception)
        self.expected = False if exception is None else not unless
        self.match = match

    def __enter__(self) -> Self:
        return self

    def _check_match(self, actual: BaseException) -> None:
        __tracebackhide__ = True
        if self.match is not None and not re.search(self.match, str(actual)):
            raise AssertionError(
                f'Pattern did not match:\n{self.match!r} (expected)\n{str(actual)!r} (actual)'
            )

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            actual: BaseException | None,
            traceback: TracebackType | None,
    ) -> bool:
        __tracebackhide__ = True
        self.raised = actual or NoException()  # type: ignore[assignment]
        if self.expected:
            if self.exception:
                actual_: type[BaseException] | BaseException | None = actual
                if actual is not None:
                    if isinstance(self.exception, type):
                        actual_ = type(actual)
                        if self.exception is not actual_:
                            return False
                        self._check_match(actual)
                    else:
                        if type(self.exception) is not type(actual):
                            return False
                compare(self.exception,
                        actual_,
                        x_label='expected',
                        y_label='raised')
            elif not actual:
                raise NoException()
            else:
                self._check_match(actual)
        elif actual:
            return False
        return True


P = ParamSpec("P")
T = TypeVar("T")


class should_raise(Generic[E]):
    __doc__ = """
    A decorator to assert that the decorated function will raised
    an exception. An exception class or exception instance may be
    passed to check more specifically exactly what exception will be
    raised.
    """ + param_docs

    @overload
    def __init__(self, exception: type[E], *, match: str | re.Pattern[str] | None = None, unless: bool | None = None) -> None: ...
    @overload
    def __init__(self, exception: singleton = not_there, *, match: str | re.Pattern[str] | None = None, unless: bool | None = None) -> None: ...
    @overload
    def __init__(self, exception: E | None, *, match: None = None, unless: bool | None = None) -> None: ...

    def __init__(
            self,
            exception: E | type[E] | None | singleton = not_there,
            *,
            match: str | re.Pattern[str] | None = None,
            unless: bool | None = None,
    ) -> None:
        self.exception: E | type[E] | None | singleton = exception
        self.match = match
        self.unless = unless

    def __call__(self, target: Callable[P, T]) -> Callable[P, None]:

        @wraps(target)
        def _should_raise_wrapper(*args: P.args, **kw: P.kwargs) -> None:
            with ShouldRaise(self.exception, match=self.match, unless=self.unless):  # type: ignore[arg-type]
                target(*args, **kw)

        return _should_raise_wrapper


@contextmanager
def ShouldAssert(expected_text: str, show_whitespace: bool = False) -> Iterator[None]:
    """
    A context manager to check that an :class:`AssertionError`
    is raised and its text is as expected.

    :param show_whitespace: If `True`, then whitespace characters in
                            multi-line strings will be replaced with their
                            representations.
    """
    __tracebackhide__ = True
    try:
        yield
    except AssertionError as e:
        actual_text = str(e)
        if expected_text != actual_text:
            if show_whitespace:
                expected_text = split_repr(expected_text)
                actual_text = split_repr(actual_text)
            raise AssertionError(diff(expected_text, actual_text,
                                      x_label='expected', y_label='actual'))
    else:
        raise AssertionError('Expected AssertionError(%r), None raised!' %
                             expected_text)



"""
Tools for helping to test Twisted applications.
"""
from typing import Callable, Any, Self, TypeAlias

import zope.interface
from constantly import NamedConstant
from twisted.logger import globalLogPublisher, formatEvent, LogLevel, ILogObserver, LogEvent
from twisted.trial.unittest import TestCase

from . import compare
from .logcapture import Entry, LogCapture as _LogCapture, build_actual_extractor, AttributeSpec

LEVEL_MAP: dict[NamedConstant, int] = {
    LogLevel.debug: 10,
    LogLevel.info: 20,
    LogLevel.warn: 30,
    LogLevel.error: 40,
    LogLevel.critical: 50,
}

LEVEL_NAME_MAP: dict[str, int] = {level.name: numeric for level, numeric in LEVEL_MAP.items()}


def level_name(event: LogEvent) -> str:
    return event['log_level'].name.upper()

LogEventAttributes: TypeAlias = AttributeSpec[LogEvent]

@zope.interface.implementer(ILogObserver)
class TwistedSource:
    """
    A :class:`~testfixtures.logcapture.CaptureSource` for Twisted log events,
    for use with :class:`~testfixtures.LogCapture`.

    On :meth:`~testfixtures.logcapture.CaptureSource.install` all existing observers on
    ``twisted.logger.globalLogPublisher`` are replaced with this source; on
    :meth:`~testfixtures.logcapture.CaptureSource.uninstall` the original observers are
    restored.

    :param attributes:
        The sequence of attributes to return for each :class:`!LogEvent` or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, a key of that name will
        be obtained from the :class:`!LogEvent`. If the key is missing, ``None`` will be used in its
        place.

        If a callable is passed, it will be called with the :class:`!LogEvent` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.

    :param level:
        The minimum log level to capture.
        Accepts an :class:`int`, a Twisted :class:`!LogLevel` constant, or a level name
        string such as ``'warn'``. The module-level constants :data:`DEBUG`, :data:`INFO`,
        :data:`WARN`, :data:`ERROR`, :data:`CRITICAL` can also be used.

        Defaults to ``0`` (capture everything).
    """

    def __init__(
        self,
        attributes: LogEventAttributes = (level_name, formatEvent),
        level: int | str | NamedConstant = 0,
    ) -> None:
        self.attributes = attributes
        if isinstance(level, str):
            self.level: int = LEVEL_NAME_MAP[level.lower()]
        elif isinstance(level, int):
            self.level = level
        else:
            self.level = LEVEL_MAP[level]
        self._collector: Callable[[Entry], None] | None = None
        self._original_observers: list | None = None
        self._compute_actual = build_actual_extractor(attributes, self.extract_field)

    def extract_field(self, raw: LogEvent, attribute: str) -> Any:
        return raw.get(attribute)

    def __call__(self, event: LogEvent) -> None:
        if self._collector is not None:
            if self.level:
                event_level = LEVEL_MAP.get(event.get('log_level'))
                if event_level is None or event_level < self.level:
                    return
            failure = event.get('log_failure')
            entry = Entry(
                raw=event,
                actual=self._compute_actual(event),
                level=LEVEL_MAP.get(event.get('log_level')),
                exception=failure.value if failure is not None else None,
            )
            self._collector(entry)

    def install(self, collector: Callable[[Entry], None]) -> None:
        self._collector = collector
        self._original_observers = globalLogPublisher._observers
        globalLogPublisher._observers = [self]

    def uninstall(self) -> None:
        if self._original_observers is not None:
            globalLogPublisher._observers = self._original_observers
            self._original_observers = None


DEFAULT_ATTRIBUTES = ('log_level', formatEvent)


class LogCapture(_LogCapture):
    """
    A helper for capturing stuff logged using Twisted's loggers.

    :param fields:
        The sequence of fields to return for each :class:`!LogEvent` or a callable that extracts
        :attr:`~testfixtures.logcapture.Entry.actual` from a record.

        If a sequence of attribute names is passed, for each item, a key of that name will
        be obtained from the :class:`!LogEvent`. If the key is missing, ``None`` will be used in its
        place.

        If a callable is passed, it will be called with the :class:`!LogEvent` and the
        value returned will be used as :attr:`~testfixtures.logcapture.Entry.actual`.
    """

    def __init__(
            self,
            fields: LogEventAttributes = DEFAULT_ATTRIBUTES,
            install: bool = False
    ) -> None:
        super().__init__(TwistedSource(fields), install=install)

    @property
    def events(self) -> list[LogEvent]:
        """The list of raw Twisted log events captured."""
        return [e.raw for e in self.entries]

    def check_failure_text(self, expected: str, index: int = -1, attribute: str = 'value') -> None:
        """
        Check the string representation of an attribute of a logged ``Failure`` is as expected.

        :param expected: The expected string representation.
        :param index: The index into :attr:`events` where the failure should have been logged.
        :param attribute: The attribute of the failure of which to find the string representation.
        """
        compare(expected, actual=str(getattr(self.events[index]['log_failure'], attribute)))

    def raise_logged_failure(self, start_index: int = 0) -> None:
        """
        A debugging tool that raises the first failure encountered in captured logging.

        :param start_index: The index into :attr:`events` from where to start looking for failures.
        """
        for event in self.events[start_index:]:
            failure = event.get('log_failure')
            if failure:
                raise failure

    @classmethod
    def make(
            cls,
            testcase: TestCase,
            fields: LogEventAttributes = DEFAULT_ATTRIBUTES
    ) -> Self:
        """
        Instantiate, install and add a cleanup for a :class:`LogCapture`.

        :param testcase: This must be an instance of :class:`twisted.trial.unittest.TestCase`.

        Any other parameters are passed directly to the :class:`LogCapture` constructor.

        :return: The :class:`LogCapture` instantiated by this method.
        """
        capture = cls(fields)
        capture.install()
        testcase.addCleanup(capture.uninstall)
        return capture


#: Short reference to Twisted's ``LogLevel.debug``
DEBUG: NamedConstant = LogLevel.debug
#: Short reference to Twisted's ``LogLevel.info``
INFO: NamedConstant = LogLevel.info
#: Short reference to Twisted's ``LogLevel.warn``
WARN: NamedConstant = LogLevel.warn
#: Short reference to Twisted's ``LogLevel.error``
ERROR: NamedConstant = LogLevel.error
#: Short reference to Twisted's ``LogLevel.critical``
CRITICAL: NamedConstant = LogLevel.critical



"""
Format handlers for serialization and deserialization of common data formats.

This module provides a consistent interface for working with JSON, YAML, and TOML formats.
Each format handler implements the Format protocol with parse() and render() methods.
"""

import json
import tomllib
from types import ModuleType
from typing import Protocol, Any, Sequence

tomlkit: ModuleType | None
try:
    import tomlkit
except ImportError:
    tomlkit = None

yaml: ModuleType | None
try:
    import yaml
except ImportError:
    yaml = None


class Format(Protocol):
    """
    Protocol for format handlers that can parse and render data.

    This protocol defines the interface for serialization formats used with
    :class:`~testfixtures.TempDirectory`. Implement this protocol to add support
    for additional formats.
    """

    #: File extensions supported by this format (with leading dot, e.g. ('.json',))
    suffixes: Sequence[str]

    def parse(self, data: str) -> Any:
        """
        Parse a string into a Python object.

        :param data: The string to parse.
        :returns: The deserialized Python object.
        """
        ...

    def render(self, obj: Any) -> str:
        """
        Render a Python object into a string.

        :param obj: The Python object to serialize.
        :returns: The serialized string.
        """
        ...


class JSONFormat:
    """JSON format handler using the standard library json module."""

    suffixes: Sequence[str] = ('.json',)

    def parse(self, data: str) -> Any:
        """Parse JSON string into a Python object."""
        return json.loads(data)

    def render(self, obj: Any) -> str:
        """Render a Python object into a JSON string."""
        return json.dumps(obj)


class YAMLFormat:
    """YAML format handler. Requires pyyaml to be installed."""

    suffixes: Sequence[str] = ('.yaml', '.yml')

    def parse(self, data: str) -> Any:
        """Parse YAML string into a Python object."""
        if yaml is None:
            raise ImportError("YAML support requires pyyaml to be installed")
        return yaml.safe_load(data)

    def render(self, obj: Any) -> str:
        """Render a Python object into a YAML string."""
        if yaml is None:
            raise ImportError("YAML support requires pyyaml to be installed")
        return yaml.safe_dump(obj)


class TOMLFormat:
    """TOML format handler. Reading uses the standard library tomllib (3.11+) or tomlkit if available. Writing requires tomlkit."""

    suffixes: Sequence[str] = ('.toml',)

    def parse(self, data: str) -> Any:
        """Parse TOML string into a Python dict."""
        if tomlkit is not None:
            return tomlkit.loads(data)
        return tomllib.loads(data)

    def render(self, obj: Any) -> str:
        """Render a Python dict into a TOML string."""
        if tomlkit is None:
            raise ImportError("TOML writing requires tomlkit to be installed")
        return tomlkit.dumps(obj)


#: JSON format handler using the standard library. Always available.
JSON: Format = JSONFormat()

#: YAML format handler. Requires the ``pyyaml`` package to be installed.
#:
#: To install the required package, use your package manager to install ``pyyaml``.
YAML: Format = YAMLFormat()

#: TOML format handler. Reading uses the standard library ``tomllib`` module
#: (Python 3.11+) or ``tomlkit`` if available. Writing requires the ``tomlkit``
#: package.
#:
#: To install the package for writing support, use your package manager to install
#: ``tomlkit``, or install testfixtures with the ``toml`` extra.
TOML: Format = TOMLFormat()



class singleton:

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return '<%s>' % self.name

    __str__ = __repr__


not_there: singleton = singleton('not_there')


from testfixtures.comparers import diff, safe_pformat, safe_repr
from testfixtures.comparison import (
    Comparison, StringComparison, RoundComparison, compare, RangeComparison,
    SequenceComparison, Subset, Permutation, MappingComparison, like, sequence,
    contains, unordered, register
)
from testfixtures.datetime import mock_datetime, mock_date, mock_time
from testfixtures.logcapture import LogCapture, log_capture, LoggingSource
from testfixtures.outputcapture import OutputCapture
from testfixtures.resolve import resolve
from testfixtures.replace import (
    Replacer,
    Replace,
    replace,
    replace_in_environ,
    replace_on_class,
    replace_in_module,
)
from testfixtures.shouldraise import ShouldRaise, should_raise, ShouldAssert
from testfixtures.shouldwarn import ShouldWarn, ShouldNotWarn
from testfixtures.tempdirectory import TempDir, TempDirectory, tempdir
from testfixtures.utils import wrap, generator


# backwards compatibility for the old names
test_datetime = mock_datetime
test_datetime.__test__ = False  # type: ignore[attr-defined]
test_date = mock_date
test_date.__test__ = False  # type: ignore[attr-defined]
test_time = mock_time
test_time.__test__ = False  # type: ignore[attr-defined]

__all__ = [
    'Comparison',
    'LogCapture',
    'LoggingSource',
    'MappingComparison',
    'OutputCapture',
    'Permutation',
    'RangeComparison',
    'Replace',
    'Replacer',
    'RoundComparison',
    'SequenceComparison',
    'ShouldAssert',
    'ShouldRaise',
    'ShouldNotWarn',
    'ShouldWarn',
    'Subset',
    'StringComparison',
    'TempDirectory',
    'TempDir',
    'compare',
    'contains',
    'diff',
    'generator',
    'like',
    'log_capture',
    'mock_date',
    'mock_datetime',
    'mock_time',
    'not_there',
    'register',
    'replace',
    'replace_in_environ',
    'replace_on_class',
    'replace_in_module',
    'resolve',
    'safe_pformat',
    'safe_repr',
    'sequence',
    'should_raise',
    'singleton',
    'tempdir',
    'test_date',
    'test_datetime',
    'test_time',
    'unordered',
    'wrap',
]
